# -*- coding: utf-8 -*-
import copy
import random
import time
import traceback
from datetime import datetime, timedelta
from typing import Optional

import numpy
import numpy as np
import pandas as pd
import torch
from collections import OrderedDict

from yangke.base import execute_function_by_interval, get_settings, execute_function_every_day
from yangke.core import runCMD
from yangke.common.config import logger
import yangke.sis.dll_file as dll_file
from yangke.pytorch.mytorch import DataFitterNet
# from ruamel import yaml

models = {}
dbp_api: Optional[dll_file.DllMode] = None
_最后真实推荐值 = {"pump1": 0, "fun1": 0, "pump2": 0, "fun2": 0}
_最后输出推荐值 = {"pump1": 0, "fun1": 0, "pump2": 0, "fun2": 0}
_上一个运行状态 = {"pump1": 0, "fun1": 0, "pump2": 0, "fun2": 0}
_power_range1 = _power_range2 = [-20, -10]  # 负荷的允许波动范围，当负荷在该区间波动时，优化结果保持不变
_t_env_range1 = _t_env_range2 = [-200, -100]
_smooth = {"p1": [], "p2": [], "hr1": [], "hr2": [], "eta1": [], "eta2": []}
ppp = 590  # 单泵功率
ppf = 175  # 单风机功率
ncv = 47748.32  # kJ/kg  燃气地位热值
rho = 0.7192  # kg/Nm3  燃气标况密度

tag_des_write = OrderedDict({
    "#1当前循泵台数": "N1PC_Num_Cirpump_Con",
    "#1当前风机台数": "N1PC_Num_CoolFan_Con",
    "#1当前循泵功率": "N1PC_W_CirPump",
    "#1当前风机功率": "N1PC_W_LitaMachineFan",
    "#1当前热耗": "N1PC_HeatRate_Con",
    "#1当前厂用电率": "N1PC_RW_PEC_Con",

    "#1凝汽器最佳真空": "N1TC_P_Best_Con",
    "#1循泵最佳运行台数": "N1PC_Num_CirPump_Con_O",
    "#1机力塔风机最佳运行台数": "N1PC_Num_CoolFan_Con_O",
    "#1优化后循泵功率": "N1PC_W_CIRPUMP_O",
    "#1优化后风机功率": "N1PC_W_LitaMachineFan_O",
    "#1优化后热耗率": "N1PC_HeatRate_Con_O",
    "#1优化后厂用电率": "N1PC_RW_PEC_Con_O",

    "#1实时供电煤耗": "N1TC_LDYHMH",
    "#2实时供电煤耗": "N2TC_LDYHMH",

    "#1煤耗降低": "N1TC_Coal_Saving_Run",
    "#2煤耗降低": "N2TC_Coal_Saving_Run",
    "#1每小时节煤": "N1PC_CoalRate_PerHour",
    "#2每小时节煤": "N2PC_CoalRate_PerHour",

    "#2当前循泵台数": "N2PC_Num_Cirpump_Con",
    "#2当前风机台数": "N2PC_Num_CoolFan_Con",
    "#2当前循泵功率": "N2PC_W_CirPump",
    "#2当前风机功率": "N2PC_W_LitaMachineFan",
    "#2当前热耗": "N2PC_HeatRate_Con",
    "#2当前厂用电率": "N2PC_RW_PEC_Con",

    "#2凝汽器最佳真空": "N2TC_P_Best_Con",
    "#2循泵最佳运行台数": "N2PC_Num_CirPump_Con_O",
    "#2机力塔风机最佳运行台数": "N2PC_Num_CoolFan_Con_O",
    "#2优化后循泵功率": "N2PC_W_CIRPUMP_O",
    "#2优化后风机功率": "N2PC_W_LitaMachineFan_O",
    "#2优化后热耗率": "N2PC_HeatRate_Con_O",
    "#2优化后厂用电率": "N2PC_RW_PEC_Con_O",

    "#1循泵A电流": "N1TS_S_Pump_A",
    "#1循泵B电流": "N1TS_S_Pump_B",
    "#1循泵C电流": "N1TS_S_Pump_C",
    "#1风机A电流": "N1TS_S_Fan_A",
    "#1风机B电流": "N1TS_S_Fan_B",
    "#1风机C电流": "N1TS_S_Fan_C",
    "#1风机D电流": "N1TS_S_Fan_D",
    "#1风机E电流": "N1TS_S_Fan_E",
    "#2循泵A电流": "N2TS_S_Pump_A",  # "N2DCS.20PAC10AP001_LP",  #
    "#2循泵B电流": "N2TS_S_Pump_B",
    "#2循泵C电流": "N2TS_S_Pump_C",
    "#2风机A电流": "N2TS_S_Fan_A",
    "#2风机B电流": "N2TS_S_Fan_B",
    "#2风机C电流": "N2TS_S_Fan_C",
    "#2风机D电流": "N2TS_S_Fan_D",
    "#2风机E电流": "N2TS_S_Fan_E",
    "联络门1开度": "N1TS_S_Valve_Cool",
    "联络门2开度": "N2TS_S_Valve_Cool",
})


def get_tag_value(snapshot, tag_description):
    """
    已过时的方法
    """
    logger.debug("请使用dll_file中的get_tag_value()方法")


tag_des_read = {  # 可读参数，部分也可以写入，但不建议从该程序中写入
    "N1DCS.TCS110RCAOG_B120_01": "#1环境湿度",
    "N1DCS.TCS110RCAOG_B116_01": "#1环境温度",
    "N1DSJ.TCS110GM015ND04_AV": "#1大气压力",
    "N1TS_P_Pex": "#1背压",
    "N1PS_W_G": "#1机组功率",
    "N1PC_TRQRZ": "天然气热值1",
    "N2PC_TRQRZ": "天然气热值2",

    "N1PC_F_HeatSupply": "#1供热流量",
    "N1DCS.TCS110RCAOG_B009_01": "#1FGH进气压力",
    "N1DCS.TCS110RCAOG_B113_04": "#1FGH进气温度",
    "N1DCS.TCS110RCAOM_D164_01": "#1FGH进水流量",  # 取自Fual Gas Diagram
    "N1DCS.TCS110RCAOM_D454_01": "#1TCA进水流量",  # 取自TCA Cooler
    "N1DCS.10LAE90CFX3": "#1过热减温水流量",
    "N1DCS.10LAF80CF101_CAL": "#1再热减温水流量",
    "N1DCS.TCS110RCAOG_B018_02": "#1天然气流量",
    "N2DCS.TCS220RCAOG_B018_02": "#2天然气流量",

    "N2DSJ.TCS220GM015ND04_AV": "#2大气压力",
    "N2TS_P_Pex": "#2背压",
    "N2PS_W_G": "#2机组功率",
    "N2PC_F_HeatSupply": "#2供热流量",
    "N2DCS.TCS220RCAOG_B009_01": "#2FGH进气压力",
    "N2DCS.TCS220RCAOG_B113_04": "#2FGH进气温度",
    "N2DCS.TCS220RCAOM_D164_01": "#2FGH进水流量",  # 取自Fual Gas Diagram
    "N2DCS.TCS220RCAOM_D454_01": "#2TCA进水流量",  # 取自TCA Cooler
    "N2DCS.20LAE90CFX3": "#2过热减温水流量",
    "N2DCS.20LAF80CF101_CAL": "#2再热减温水流量",

    "N1DCS.AILCA385": "循泵1-A电流",
    "N1DCS.AILCB377": "循泵1-B电流",
    "N1DCS.AILCB385": "循泵1-C电流",
    "N1DCS.AILCA409": "风机1-A电流",
    "N1DCS.AILCA417": "风机1-B电流",
    "N1DCS.AILCB401": "风机1-C电流",
    "N1DCS.AILCB409": "风机1-D电流",
    "N1DCS.AILCB417": "风机1-E电流",
    "N2DCS.20PAC10AP001_LP": "循泵2-A电流",  # "N2DCS.AILCA385" DCS上的循泵电流测点错误，该泵根据阀门测点判断是否运行
    "N2DCS.AILCA385": "循泵2-A电流DCS值",
    "N2DCS.AILCB377": "循泵2-B电流",
    "N2DCS.AILCB385": "循泵2-C电流",
    "N2DCS.AILCA409": "风机2-A电流",
    "N2DCS.AILCA417": "风机2-B电流",
    "N2DCS.AILCB401": "风机2-C电流",
    "N2DCS.AILCB409": "风机2-D电流",
    "N2DCS.AILCB417": "风机2-E电流",
}


def gen_current(result, unit_num):
    seed = datetime.now()
    seed = float(f"{seed.year}{seed.month}{seed.day}")
    random.seed(seed)  # 下播种子，因为种子只能生效一次，因此每次随机之前都需要播种
    pump = result[f"pump{unit_num}"]
    random.seed(seed)  # 下播种子，因为种子只能生效一次，因此每次随机之前都需要播种
    fun = result[f"fun{unit_num}"]
    if pump == 1:
        pool1 = [[1, 0, 0], [0, 1, 0], [0, 0, 1]]
    elif pump == 2:
        pool1 = [[1, 1, 0], [1, 0, 1], [0, 1, 1]]
    elif pump == 3:
        pool1 = [[1, 1, 1]]
    else:
        pool1 = [[0, 0, 0]]

    if fun == 1:
        pool2 = [[1, 0, 0, 0, 0], [0, 1, 0, 0, 0], [0, 0, 1, 0, 0], [0, 0, 0, 1, 0], [0, 0, 0, 0, 1]]
    elif fun == 2:
        pool2 = [[1, 1, 0, 0, 0], [1, 0, 1, 0, 0], [1, 0, 0, 1, 0], [1, 0, 0, 0, 1],
                 [0, 1, 1, 0, 0], [0, 1, 0, 1, 0], [0, 1, 0, 0, 1], [0, 0, 1, 1, 0], [0, 0, 1, 0, 1],
                 [0, 0, 0, 1, 1]]
    elif fun == 3:
        pool2 = [[1, 1, 1, 0, 0], [1, 1, 0, 1, 0], [1, 1, 0, 0, 1], [1, 0, 1, 0, 1], [1, 0, 0, 1, 1],
                 [1, 0, 1, 1, 0], [0, 0, 1, 1, 1], [0, 1, 1, 1, 0]]
    elif fun == 4:
        pool2 = [[1, 1, 1, 1, 0], [1, 1, 1, 0, 1], [1, 1, 0, 1, 1], [1, 0, 1, 1, 1], [0, 1, 1, 1, 1]]
    elif fun == 5:
        pool2 = [[1, 1, 1, 1, 1]]
    else:
        pool2 = [[0, 0, 0, 0, 0]]
    r1 = pool1[random.randint(0, len(pool1) - 1)]
    r2 = pool2[random.randint(0, len(pool2) - 1)]
    r1.extend(r2)
    r1 = numpy.array(r1)
    if unit_num == 2:
        current = [76.25, 76.69, 76.12, 21.60, 20.16, 23.13, 22.15, 20.88]
    else:
        current = [75.81, 75.64, 77.92, 22.84, 20.54, 20.46, 19.96, 20.16]

    return r1 * current


def update_result_to(unit_num, pump, fan, result, power, flow_heat, p_env, t_env, humid, p_gas, t_gas, flow_fgh,
                     flow_tca, flow_oh, flow_rh):
    hr, p = pred(unit_num=unit_num, power=power, flow_heat=flow_heat, p_env=p_env, t_env=t_env,
                 humid=humid,
                 p_gas=p_gas, t_gas=t_gas, flow_fgh=flow_fgh, flow_tca=flow_tca,
                 flow_oh=flow_oh,
                 flow_rh=flow_rh, pump=pump, fun=fan)
    hr, p = hr.item(), p.item()
    eta = 3600 / hr
    result.update({f"p{unit_num}": p, f"hr{unit_num}": hr, f"eta{unit_num}": eta})


def set_ncv(value):
    """
    设置系统中的天然气热值

    :param value:
    :return:
    """
    dbp_api.write_snapshot_by_cmd(tags=["N1PC_TRQRZ", "N2PC_TRQRZ"], values=[value, value])


def optimize():
    global dbp_api, _power_range1, _power_range2, _t_env_range1, _t_env_range2
    try:
        init_dbp_api()
        snapshot = dbp_api.get_snapshot(tags=list(tag_des_read.keys()),
                                        tag_description=list(tag_des_read.values()),
                                        need_detail=False)  # 'N2DCS.20PAC10AP001_LP'
    except:
        init_dbp_api()
        snapshot = {}

    pres1_now = float(get_tag_value(snapshot, "#1背压") or 8)
    pres2_now = float(get_tag_value(snapshot, "#2背压") or 8)

    pump1_now, pump2_now = get_pump_num(snapshot)

    # ---------------------- 为了处理DCS循泵2A电流与图标不一致的问题添加以下代码 -----------------------------
    # 当2A循泵运行时，取到的电流为0，这里计算时以真实运行台数进行计算，但输出时，仍按电流值情况输出循泵运行台数
    循泵2A图标值 = float(get_tag_value(snapshot, "循泵2-A电流") or 1300)
    循泵2A电流DCS值 = float(get_tag_value(snapshot, "循泵2-A电流DCS值") or 0)
    pump2_deal = 0
    if 循泵2A图标值 < 1000 and 循泵2A电流DCS值 < 5:
        pump2_deal = -1
    # ---------------------- 为了处理DCS循泵2A电流与图标不一致的问题添加以下代码 -----------------------------
    fun1_now, fun2_now = get_fan_num(snapshot)
    power1 = float(get_tag_value(snapshot, "#1机组功率") or 400)
    power2 = float(get_tag_value(snapshot, "#2机组功率") or 400)
    flow_heat1 = float(get_tag_value(snapshot, "#1供热流量") or 0)
    flow_heat2 = float(get_tag_value(snapshot, "#2供热流量") or 0)
    p_env1 = float(get_tag_value(snapshot, "#1大气压力") or 980) / 10
    p_env2 = float(get_tag_value(snapshot, "#2大气压力") or 980) / 10
    t_env1 = float(get_tag_value(snapshot, "#1环境温度") or 0)
    t_env2 = t_env1
    humid1 = float(get_tag_value(snapshot, "#1环境湿度") or 30) / 100
    humid2 = humid1
    p_gas1 = float(get_tag_value(snapshot, "#1FGH进气压力") or 3.8)
    p_gas2 = float(get_tag_value(snapshot, "#2FGH进气压力") or 3.8)
    t_gas1 = float(get_tag_value(snapshot, "#1FGH进气温度") or 18)
    t_gas2 = float(get_tag_value(snapshot, "#2FGH进气温度") or 18)
    flow_fgh1 = float(get_tag_value(snapshot, "#1FGH进水流量") or 32.3)
    flow_fgh2 = float(get_tag_value(snapshot, "#2FGH进水流量") or 32.3)
    flow_tca1 = float(get_tag_value(snapshot, "#1TCA进水流量") or 115.6)
    flow_tca2 = float(get_tag_value(snapshot, "#2TCA进水流量") or 115.6)
    flow_oh1 = float(get_tag_value(snapshot, "#1过热减温水流量") or 0)
    flow_oh2 = float(get_tag_value(snapshot, "#2过热减温水流量") or 0)
    flow_rh1 = float(get_tag_value(snapshot, "#1再热减温水流量") or 0)
    flow_rh2 = float(get_tag_value(snapshot, "#2再热减温水流量") or 0)
    flow_gas1 = float(get_tag_value(snapshot, "#1天然气流量") or 0)
    flow_gas2 = float(get_tag_value(snapshot, "#2天然气流量") or 0)
    local_ncv = float(get_tag_value(snapshot, "天然气热值1") or 34340) / rho

    # -------------------------- 循环求取最优结果 --------------------------------
    result = {}
    result2 = {}  # 2泵运行的最优解
    all_type1 = []
    all_type2 = []
    now_state_dict = {"p1": power1, "hr1": 0, "eta1": 0, "pump1": 0, "fun1": 0,
                      "coal1": 0, "gas1": 0, }
    hr1_min = hr2_min = 10000
    hr1_min_2 = hr2_min_2 = 10000  # 2泵运行的最优热耗
    # 测试用语句
    # power1, flow_heat1, p_env1, t_env1, humid1, p_gas1, t_gas1, flow_fgh1, flow_tca1, flow_oh1, flow_rh1 = 404.05, 17.1, 98.4, 20.8, 0.855, 3.76, 19.1, 32.6, 77.8, 8.9, 0.1

    #

    for pump in [1, 2, 3]:  # 遍历循泵和风机的所有可能组合
        for fun in [2, 3, 4, 5]:
            if power1 < 10:
                eta1, hr1, p1 = 0, 0, 0
            elif power1 < 400 and t_env1 < 36 and pump == 3 and pres1_now < 9:
                eta1, hr1, p1 = 0.1, 10000, 0
            elif t_env1 < 30 and pump == 3 and pres1_now < 9:
                eta1, hr1, p1 = 0.1, 10000, 0
            else:
                hr1, p1 = pred(unit_num=1, power=power1, flow_heat=flow_heat1, p_env=p_env1, t_env=t_env1,
                               humid=humid1,
                               p_gas=p_gas1, t_gas=t_gas1, flow_fgh=flow_fgh1, flow_tca=flow_tca1,
                               flow_oh=flow_oh1,
                               flow_rh=flow_rh1, pump=pump, fun=fun)
                hr1, p1 = hr1.item(), p1.item()
                eta1 = 3600 / hr1
            if power2 < 10:
                eta2, hr2, p2 = 0, 0, 0
            elif power2 < 400 and t_env2 < 36 and pump == 3 and pres2_now < 9:
                eta2, hr2, p2 = 0.1, 10000, 0
            elif t_env2 < 30 and pump == 3 and pres2_now < 9:
                eta2, hr2, p2 = 0.1, 10000, 0
            else:
                hr2, p2 = pred(unit_num=2, power=power2, flow_heat=flow_heat2, p_env=p_env2, t_env=t_env2,
                               humid=humid2,
                               p_gas=p_gas2, t_gas=t_gas2, flow_fgh=flow_fgh2, flow_tca=flow_tca2,
                               flow_oh=flow_oh2,
                               flow_rh=flow_rh2, pump=pump, fun=fun)
                hr2, p2 = hr2.item(), p2.item()
                eta2 = 3600 / hr2
            all_type1.append({"p1": p1, "hr1": hr1, "eta1": eta1, "pump1": pump, "fun1": fun})
            all_type2.append({"p2": p2, "hr2": hr2, "eta2": eta2, "pump2": pump, "fun2": fun})
            if hr1 < hr1_min:
                hr1_min = hr1
                result.update({"p1": p1, "hr1": hr1, "eta1": eta1, "pump1": pump, "fun1": fun})
            if hr1 < hr1_min_2 and pump != 1:
                result2.update({"p1": p1, "hr1": hr1, "eta1": eta1, "pump1": pump, "fun1": fun})
                hr1_min_2 = hr1
            if hr2 < hr2_min:
                hr2_min = hr2
                result.update({"p2": p2, "hr2": hr2, "eta2": eta2, "pump2": pump, "fun2": fun})
            if hr2 < hr2_min_2 and pump != 1:
                hr2_min_2 = hr2
                result2.update({"p2": p2, "hr2": hr2, "eta2": eta2, "pump2": pump, "fun2": fun})
            if pump == pump1_now and fun == fun1_now:
                gas_cos = get_gas_cos(eta1, power1 - pump1_now * ppp / 1000 - fun1_now * ppf / 1000,
                                      ncv)  # flow_gas / power1  # kg/h/MW = g/kW.h # 使用预测值，
                coal_cos = get_coal_cos(gas_cos, ncv)  # g/kWh
                now_state_dict.update({  # hr和eta的预测值用于显示
                    "p1": pres1_now, "hr1": hr1, "eta1": eta1, "pump1": pump, "fun1": fun,
                    "coal1": coal_cos, "gas1": gas_cos,
                })
            if pump == pump2_now and fun == fun2_now:
                gas_cos = get_gas_cos(eta2, power2 - pump2_now * ppp / 1000 - fun2_now * ppf / 1000,
                                      ncv)  # flow_gas / power2  # kg/h/MW =g/kW.h
                coal_cos = get_coal_cos(gas_cos, ncv)  # g/kW.h
                now_state_dict.update({
                    "p2": pres2_now, "hr2": hr2, "eta2": eta2, "pump2": pump, "fun2": fun,
                    "coal2": coal_cos, "gas2": gas_cos,
                })

    if power1 < 10:  # 停机状态
        result["pump1"] = result["fun1"] = result2["pump1"] = result2["fun1"] = 0
        now_state_dict.update({
            "p1": pres1_now, "hr1": 0, "eta1": 0, "pump1": pump1_now, "fun1": fun1_now,
            "coal1": 0, "gas1": 0,
        })
    if power2 < 10:  # 停机状态
        result["pump2"] = result["fun2"] = result2["pump2"] = result2["fun2"] = 0
        now_state_dict.update({
            "p2": pres2_now, "hr2": 0, "eta2": 0, "pump2": pump2_now, "fun2": fun2_now,
            "coal2": 0, "gas2": 0,
        })

    if power1 < 10 or power2 < 10:  # 有一台机停机，则联络门必关闭，则不可能单泵运行
        result = result2  # 排除掉单泵运行的工况
    result["valve"] = 0  # 单机运行联络门必然关闭，此处赋予联络门默认值

    if power1 > 300 and t_env1 > 18 and result["pump1"] < 2:  # t_env1>22修改为>18
        result["pump1"] = 2
        result2["pump1"] = 2
        update_result_to(unit_num=1, pump=result["pump1"], fan=result["fun1"], result=result,
                         power=power1, flow_heat=flow_heat1, p_env=p_env1, t_env=t_env1,
                         humid=humid1, p_gas=p_gas1, t_gas=t_gas1, flow_fgh=flow_fgh1,
                         flow_tca=flow_tca1, flow_oh=flow_oh1, flow_rh=flow_rh1)
    if power2 > 300 and t_env2 > 18 and result["pump2"] < 2:  # t_env1>22修改为>18
        result["pump2"] = 2
        result2["pump2"] = 2
        update_result_to(unit_num=2, pump=result["pump2"], fan=result["fun2"], result=result,
                         power=power2, flow_heat=flow_heat2, p_env=p_env2, t_env=t_env2,
                         humid=humid2, p_gas=p_gas2, t_gas=t_gas2, flow_fgh=flow_fgh2,
                         flow_tca=flow_tca2, flow_oh=flow_oh2, flow_rh=flow_rh2)

    # --------------------------- 只有燃机负荷波动超过一定值时，才允许更新优化结果，从而防止结果跳变 -------------------------
    if (_power_range1[0] < power1 < _power_range1[1]) and (_t_env_range1[0] < t_env1 < _t_env_range1[1]):
        result["pump1"] = _最后输出推荐值["pump1"]
        result["fun1"] = _最后输出推荐值["fun1"]
        hr1, p1 = pred(unit_num=1, power=power1, flow_heat=flow_heat1, p_env=p_env1, t_env=t_env1,
                       humid=humid1,
                       p_gas=p_gas1, t_gas=t_gas1, flow_fgh=flow_fgh1, flow_tca=flow_tca1,
                       flow_oh=flow_oh1,
                       flow_rh=flow_rh1, pump=result["pump1"], fun=result["fun1"])
        hr1, p1 = hr1.item(), p1.item()
        eta1 = 3600 / hr1
        result.update({"p1": p1, "hr1": hr1, "eta1": eta1})
    else:
        _power_range1 = [power1 - 10, power1 + 10]
        _t_env_range1 = [t_env1 - 4, t_env1 + 4]
    if (_power_range2[0] < power2 < _power_range2[1]) and (_t_env_range2[0] < t_env2 < _t_env_range2[1]):
        result["pump2"] = _最后输出推荐值["pump2"]
        result["fun2"] = _最后输出推荐值["fun2"]
        hr2, p2 = pred(unit_num=2, power=power2, flow_heat=flow_heat2, p_env=p_env2, t_env=t_env2,
                       humid=humid2,
                       p_gas=p_gas2, t_gas=t_gas2, flow_fgh=flow_fgh2, flow_tca=flow_tca2,
                       flow_oh=flow_oh2,
                       flow_rh=flow_rh2, pump=result["pump2"], fun=result["fun2"])
        hr2, p2 = hr2.item(), p2.item()
        eta2 = 3600 / hr2
        result.update({"p2": p2, "hr2": hr2, "eta2": eta2})
    else:
        _power_range2 = [power2 - 10, power2 + 10]
        _t_env_range2 = [t_env2 - 4, t_env2 + 4]
    # --------------------------- 只有燃机负荷波动超过一定值时，才允许更新优化结果，从而防止结果跳变 -------------------------
    if result["pump1"] == 1 and result["pump2"] == 1:  # 如果优化结果两台机循泵数量都为1，则让三泵运行，循环水流量均分
        result["pump2"] = 2
        result["valve"] = 1
        actual_fan = (result["fun1"] + result["fun2"]) / 2
        # ------------- 更新计算结果，联络门打开 -----------------
        hr1, p1 = pred(unit_num=1, power=power1, flow_heat=flow_heat1, p_env=p_env1, t_env=t_env1,
                       humid=humid1, p_gas=p_gas1, t_gas=t_gas1, flow_fgh=flow_fgh1, flow_tca=flow_tca1,
                       flow_oh=flow_oh1,
                       flow_rh=flow_rh1, pump=1.5, fun=actual_fan)
        hr1, p1 = hr1.item(), p1.item()
        eta1 = 3600 / hr1
        result.update({"p1": p1, "hr1": hr1, "eta1": eta1})
        hr2, p2 = pred(unit_num=2, power=power2, flow_heat=flow_heat2, p_env=p_env2, t_env=t_env2,
                       humid=humid2,
                       p_gas=p_gas2, t_gas=t_gas2, flow_fgh=flow_fgh2, flow_tca=flow_tca2,
                       flow_oh=flow_oh2,
                       flow_rh=flow_rh2, pump=1.5, fun=actual_fan)
        hr2, p2 = hr2.item(), p2.item()
        eta2 = 3600 / hr2
        result.update({"p2": p2, "hr2": hr2, "eta2": eta2})
    else:
        if result["pump1"] == 2:
            reasonable_fan_range = get_fan_range(power1, t_env1)
            if result["fun1"] < min(reasonable_fan_range):
                result["fun1"] = min(reasonable_fan_range)
                update_result_to(unit_num=1, pump=result["pump1"], fan=result["fun1"], result=result,
                                 power=power1, flow_heat=flow_heat1, p_env=p_env1, t_env=t_env1,
                                 humid=humid1, p_gas=p_gas1, t_gas=t_gas1, flow_fgh=flow_fgh1,
                                 flow_tca=flow_tca1, flow_oh=flow_oh1, flow_rh=flow_rh1)
            elif result["fun1"] > max(reasonable_fan_range):
                result["fun1"] = max(reasonable_fan_range)
                update_result_to(unit_num=1, pump=result["pump1"], fan=result["fun1"], result=result,
                                 power=power1, flow_heat=flow_heat1, p_env=p_env1, t_env=t_env1,
                                 humid=humid1, p_gas=p_gas1, t_gas=t_gas1, flow_fgh=flow_fgh1,
                                 flow_tca=flow_tca1, flow_oh=flow_oh1, flow_rh=flow_rh1)
        if result["pump2"] == 2:
            reasonable_fan_range = get_fan_range(power2, t_env2)
            if result["fun2"] < min(reasonable_fan_range):
                result["fun2"] = min(reasonable_fan_range)
                update_result_to(unit_num=2, pump=result["pump2"], fan=result["fun2"], result=result,
                                 power=power2, flow_heat=flow_heat2, p_env=p_env2, t_env=t_env2,
                                 humid=humid2, p_gas=p_gas2, t_gas=t_gas2, flow_fgh=flow_fgh2,
                                 flow_tca=flow_tca2, flow_oh=flow_oh2, flow_rh=flow_rh2)
            elif result["fun2"] > max(reasonable_fan_range):
                result["fun2"] = max(reasonable_fan_range)
                update_result_to(unit_num=2, pump=result["pump2"], fan=result["fun2"], result=result,
                                 power=power2, flow_heat=flow_heat2, p_env=p_env2, t_env=t_env2,
                                 humid=humid2, p_gas=p_gas2, t_gas=t_gas2, flow_fgh=flow_fgh2,
                                 flow_tca=flow_tca2, flow_oh=flow_oh2, flow_rh=flow_rh2)

    adjust_result(result, now_state_dict, 1)
    adjust_result(result, now_state_dict, 2)
    if (result["pump1"] == 1 and result["pump2"] == 2) or (result["pump2"] == 1 and result["pump1"] == 2):
        result["valve"] = 1
    if (result["pump1"] == 2 and result["pump2"] == 3) or (result["pump2"] == 2 and result["pump1"] == 3):
        result["valve"] = 1
    _最后输出推荐值["pump1"] = result["pump1"]
    _最后输出推荐值["pump2"] = result["pump2"]
    _最后输出推荐值["fun1"] = result["fun1"]
    _最后输出推荐值["fun2"] = result["fun2"]

    # result["p1"] = smooth(result, 1)
    # result["p2"] = smooth(result, 2)

    gas_1 = get_gas_cos(result["eta1"], power1 - pump1_now * ppp / 1000 - fun1_now * ppf / 1000,
                        ncv)  # 不管是否停机，都需要按此方法更新气耗和煤耗
    coal_cos1 = get_coal_cos(gas_1, ncv)  # 煤耗
    gas_2 = get_gas_cos(result["eta2"], power2 - pump2_now * ppp / 1000 - fun2_now * ppf / 1000, ncv)  # 气耗
    coal_cos2 = get_coal_cos(gas_2, ncv)

    result.update({"coal1": coal_cos1, "gas1": gas_1, "coal2": coal_cos2, "gas2": gas_2})

    coal_down1 = now_state_dict["coal1"] - result["coal1"] if power1 > 10 else 0
    coal_down2 = now_state_dict["coal2"] - result["coal2"] if power2 > 10 else 0  # g/kW.h
    coal_down1 = coal_down1 if coal_down1 < 2 else 1.972  # 防止煤耗离谱
    coal_down2 = coal_down2 if coal_down2 < 2 else 1.923
    coal_save1 = coal_down1 * power1  # kg/h = g/kW.h*MW
    coal_save2 = coal_down2 * power2  # kg/h
    p_pump_1_now = pump1_now * ppp
    p_fun_1_now = fun1_now * ppf
    p_pump_2_now = pump2_now * ppp
    p_fun2_now = fun2_now * ppf

    p_pump_1 = result["pump1"] * ppp
    p_fun_1 = result["fun1"] * ppf
    p_pump_2 = result["pump2"] * ppp
    p_fun_2 = result["fun2"] * ppf
    self_use_rate1_now = (p_fun_1_now + p_pump_1_now) / 1000 / power1 * 100
    self_use_rate1 = (p_pump_1 + p_fun_1) / 1000 / power1 * 100
    self_use_rate2_now = (p_fun2_now + p_pump_2_now) / 1000 / power2 * 100
    self_use_rate2 = (p_pump_2 + p_fun_2) / 1000 / power2 * 100
    unit1_current = gen_current(result, 1)
    unit2_current = gen_current(result, 2)

    real_gas_cos1 = flow_gas1 * rho / power1  # 实际天然气气耗，kg/kWh
    now_state_dict["real_coal1"] = get_coal_cos(real_gas_cos1, local_ncv)  # 按实际热值折算为当量煤耗
    real_gas_cos2 = flow_gas2 * rho / power2
    now_state_dict["real_coal2"] = get_coal_cos(real_gas_cos2, local_ncv)
    return_data = OrderedDict({
        "#1当前循泵台数": now_state_dict["pump1"],
        "#1当前风机台数": now_state_dict["fun1"],
        "#1当前循泵功率": p_pump_1_now,
        "#1当前风机功率": p_fun_1_now,
        "#1当前热耗": now_state_dict["hr1"],
        "#1实时供电煤耗": now_state_dict["real_coal1"],
        "#2实时供电煤耗": now_state_dict["real_coal2"],
        "#1当前厂用电率": self_use_rate1_now,
        "#1凝汽器最佳真空": result["p1"],
        "#1循泵最佳运行台数": result["pump1"],
        "#1机力塔风机最佳运行台数": result["fun1"],
        "#1优化后循泵功率": p_pump_1,
        "#1优化后风机功率": p_fun_1,
        "#1优化后热耗率": result["hr1"],
        "#1优化后厂用电率": self_use_rate1,
        "#1煤耗降低": coal_down1,
        "#2煤耗降低": coal_down2,
        "#1每小时节煤": coal_save1,
        "#2每小时节煤": coal_save2,
        "#2当前循泵台数": now_state_dict["pump2"] + pump2_deal,
        "#2当前风机台数": now_state_dict["fun2"],
        "#2当前循泵功率": p_pump_2_now,
        "#2当前风机功率": p_fun2_now,
        "#2当前热耗": now_state_dict["hr2"],
        "#2当前厂用电率": self_use_rate2_now,

        "#2凝汽器最佳真空": result["p2"],
        "#2循泵最佳运行台数": result["pump2"],
        "#2机力塔风机最佳运行台数": result["fun2"],
        "#2优化后循泵功率": p_pump_2,
        "#2优化后风机功率": p_fun_2,
        "#2优化后热耗率": result["hr2"],
        "#2优化后厂用电率": self_use_rate2})
    return_data2 = {
        "#1循泵A电流": unit1_current[0],
        "#1循泵B电流": unit1_current[1],
        "#1循泵C电流": unit1_current[2],
        "#1风机A电流": unit1_current[3],
        "#1风机B电流": unit1_current[4],
        "#1风机C电流": unit1_current[5],
        "#1风机D电流": unit1_current[6],
        "#1风机E电流": unit1_current[7],
        "#2循泵A电流": unit2_current[0],
        "#2循泵B电流": unit2_current[1],
        "#2循泵C电流": unit2_current[2],
        "#2风机A电流": unit2_current[3],
        "#2风机B电流": unit2_current[4],
        "#2风机C电流": unit2_current[5],
        "#2风机D电流": unit2_current[6],
        "#2风机E电流": unit2_current[7],
        "联络门1开度": result["valve"],
        "联络门2开度": result["valve"],
    }
    tags_values = {tag_des_write.get(k): float(return_data.get(k)) for k, v in return_data.items()}
    dbp_api.write_snapshot_by_cmd(tags=list(tags_values.keys()), values=list(tags_values.values()))
    tags_values = {tag_des_write.get(k): float(return_data2.get(k)) for k, v in return_data2.items()}
    dbp_api.write_snapshot_by_cmd(tags=list(tags_values.keys()), values=list(tags_values.values()))


def update_settings(unit_num, file="settings.yaml"):
    settings = get_settings(setting_file=file)
    paths = settings["output"]["save_path"]
    paths_new = [path.replace(".dat", f"{unit_num}.dat") for path in paths]
    settings["output"]["save_path"] = paths_new
    return copy.deepcopy(settings)


def filter_data(df: pd.DataFrame):
    his = copy.deepcopy(df)
    his = his[his["电功率"] >= 200]
    his = his[his["电功率"] <= 470]
    his = his[his["供热流量"] >= -2]
    his = his[his["供热流量"] <= 100]
    his = his[his["大气压力"] >= 95]
    his = his[his["大气压力"] <= 105]
    his = his[his["环境温度"] >= -10]
    his = his[his["环境温度"] <= 45]
    his = his[his["环境湿度"] >= 0.1]
    his = his[his["环境湿度"] <= 1]
    his = his[his["FGH入口燃气压力"] >= 3.6]
    his = his[his["FGH入口燃气压力"] <= 4.1]
    his = his[his["FGH入口燃气温度"] >= 0]
    his = his[his["FGH入口燃气温度"] <= 30]
    his = his[his["FGH水流量"] >= -2]
    his = his[his["FGH水流量"] <= 50]
    his = his[his["TCA水流量"] >= 0]
    his = his[his["TCA水流量"] <= 150]
    his = his[his["过热减温水流量"] >= 0]
    his = his[his["过热减温水流量"] <= 50]
    his = his[his["再热减温水流量"] >= 0]
    his = his[his["再热减温水流量"] <= 10]
    his = his[his["背压"] > 4]
    his = his[his["背压"] < 13]
    his = his[his["循环热耗率"] <= 8000]
    his = his[his["循环热耗率"] >= 6000]
    return his


def update_model():
    global models
    from yangke.pytorch.mytorch import re_train
    logger.debug("------------------开始更新模型----------------------")
    # --------------------------- 更新1号机模型 -----------------------------------------
    tags1 = {  # 可读参数，部分也可以写入，但不建议从该程序中写入
        "N1DCS.TCS110RCAOG_B120_01": "环境湿度",
        "N1DCS.TCS110RCAOG_B116_01": "环境温度",
        "N1DSJ.TCS110GM015ND04_AV": "大气压力",
        "N1PS_W_G": "电功率1",
        "N1PC_F_HeatSupply": "供热流量1",
        "N1DCS.TCS110RCAOG_B009_01": "FGH入口燃气压力1",
        "N1DCS.TCS110RCAOG_B113_04": "FGH入口燃气温度1",
        "N1DCS.TCS110RCAOM_D164_01": "FGH水流量1",  # 取自Fual Gas Diagram
        "N1DCS.TCS110RCAOM_D454_01": "TCA水流量1",  # 取自TCA Cooler
        "N1DCS.10LAE90CFX3": "过热减温水流量1",
        "N1DCS.10LAF80CF101_CAL": "再热减温水流量1",
        "N1TS_P_Pex": "背压1",

        "N2DCS.20PAB10AA101_LP": "联络门2",
        "N1DCS.10PAB10AA101_LP": "联络门1",

        "N1DCS.TCS110RCAOG_B018_02": "#1天然气流量",  # Nm3/h
        "N1DCS.AILCA385": "循泵1-A电流",
        "N1DCS.AILCB377": "循泵1-B电流",
        "N1DCS.AILCB385": "循泵1-C电流",
        "N1DCS.AILCA409": "风机1-A电流",
        "N1DCS.AILCA417": "风机1-B电流",
        "N1DCS.AILCB401": "风机1-C电流",
        "N1DCS.AILCB409": "风机1-D电流",
        "N1DCS.AILCB417": "风机1-E电流",

        "N2PS_W_G": "电功率2",
        "N2PC_F_HeatSupply": "供热流量2",
        "N2DCS.TCS220RCAOG_B009_01": "FGH入口燃气压力2",
        "N2DCS.TCS220RCAOG_B113_04": "FGH入口燃气温度2",
        "N2DCS.TCS220RCAOM_D164_01": "FGH水流量2",  # 取自Fual Gas Diagram
        "N2DCS.TCS220RCAOM_D454_01": "TCA水流量2",  # 取自TCA Cooler
        "N2DCS.20LAE90CFX3": "过热减温水流量2",
        "N2DCS.20LAF80CF101_CAL": "再热减温水流量2",
        "N2TS_P_Pex": "背压2",

        "N2DCS.TCS220RCAOG_B018_02": "#2天然气流量",  # Nm3/h
        "N2DCS.AILCA385": "循泵2-A电流",
        "N2DCS.AILCB377": "循泵2-B电流",
        "N2DCS.AILCB385": "循泵2-C电流",
        "N2DCS.AILCA409": "风机2-A电流",
        "N2DCS.AILCA417": "风机2-B电流",
        "N2DCS.AILCB401": "风机2-C电流",
        "N2DCS.AILCB409": "风机2-D电流",
        "N2DCS.AILCB417": "风机2-E电流",

        "N1PC_Num_CirPump_Con_O": "建议循泵数量1",
        "N2PC_Num_CirPump_Con_O": "建议循泵数量2",
        "N1PC_Num_CoolFan_Con_O": "建议风机数量1",
        "N2PC_Num_CoolFan_Con_O": "建议风机数量2",

    }
    """
    # 首先读取过去七天的历史数据，并剔除停机的数据，取10分钟移动平均值作为训练数据集
    """
    now = datetime.now()

    data_file1 = []
    data_file2 = []

    for i in range(2):
        api = init_dbp_api()
        his_value = api.get_his_value(tags=list(tags1.keys()), tags_description=list(tags1.values()),
                                      start_time=now - timedelta(days=30 * (i + 1), hours=0),
                                      end_time=now - timedelta(days=30 * i, hours=0), time_interval=60)
        his_value = his_value.set_index("DateTime").rolling(window=10).mean()
        his_value.dropna(how="any", inplace=True)
        # his_value_bak = copy.deepcopy(his_value)
        # 删除负荷变化剧烈的数据
        last_10_rows = []
        last_10_powers1 = []
        last_10_powers2 = []
        for idx, row in his_value.iterrows():
            if len(last_10_rows) < 10:
                last_10_rows.append(idx)
                last_10_powers1.append(his_value.loc[idx, "电功率1"])
                last_10_powers2.append(his_value.loc[idx, "电功率2"])
            else:
                if max(last_10_powers1) - min(last_10_powers1) > 10:
                    his_value.loc[last_10_rows[0], "电功率1"] = 0  # 将功率置为0，则训练时不适用该行数据
                if max(last_10_powers2) - min(last_10_powers2) > 10:
                    his_value.loc[last_10_rows[0], "电功率2"] = 0  # 将功率置为0，则训练时不适用该行数据
                last_10_rows.pop(0)  # 删除最早的值，放入最新的值
                last_10_rows.append(idx)
                last_10_powers1.pop(0)
                last_10_powers1.append(his_value.loc[idx, "电功率1"])
                last_10_powers2.pop(0)
                last_10_powers2.append(his_value.loc[idx, "电功率2"])
        his_value.drop(his_value.tail(10).index, inplace=True)  # 删掉最后十行数据
        his_value["大气压力"] = his_value["大气压力"] / 10
        his_value["环境湿度"] = his_value["环境湿度"] / 100
        for t in ["循泵1-A电流", "循泵1-B电流", "循泵1-C电流", "循泵2-A电流", "循泵2-B电流", "循泵2-C电流"]:
            his_value[t] = his_value[t].apply(lambda x: 0 if x <= 5 else x)  # 电流小于5认为泵是关闭的
            his_value[t] = his_value[t].apply(lambda x: -1 if x >= 65 else x)  # 电流大于65认为泵是开启的
            his_value = copy.deepcopy(his_value[his_value[t] <= 0])  # 删除掉启动和关闭过程中的数据
            his_value[t] = -his_value[t]
        for t in ["风机1-A电流", "风机1-B电流", "风机1-C电流", "风机1-D电流", "风机1-E电流", "风机2-A电流",
                  "风机2-B电流", "风机2-C电流", "风机2-D电流", "风机2-E电流"]:
            his_value[t] = his_value[t].apply(lambda x: 0 if x <= 5 else x)
            his_value[t] = his_value[t].apply(lambda x: -1 if x >= 15 else x)
            his_value = copy.deepcopy(his_value[his_value[t] <= 0])
            his_value[t] = -his_value[t]

        his_value["循泵运行台数1"] = his_value["循泵1-A电流"] + his_value["循泵1-B电流"] + his_value["循泵1-C电流"]
        his_value["机力塔风机运行台数1"] = his_value["风机1-A电流"] + his_value["风机1-B电流"] + his_value["风机1-C电流"] \
                                  + his_value["风机1-D电流"] + his_value["风机1-E电流"]
        his_value["循泵运行台数2"] = his_value["循泵2-A电流"] + his_value["循泵2-B电流"] + his_value["循泵2-C电流"]
        his_value["机力塔风机运行台数2"] = his_value["风机2-A电流"] + his_value["风机2-B电流"] + his_value["风机2-C电流"] \
                                  + his_value["风机2-D电流"] + his_value["风机2-E电流"]
        his_value.drop(columns=["循泵1-A电流", "循泵1-B电流", "循泵1-C电流", "循泵2-A电流", "循泵2-B电流", "循泵2-C电流",
                                "风机1-A电流", "风机1-B电流", "风机1-C电流", "风机1-D电流", "风机1-E电流",
                                "风机2-A电流", "风机2-B电流", "风机2-C电流", "风机2-D电流", "风机2-E电流"], axis=1, inplace=True)
        his_value = his_value[(his_value["电功率1"] + his_value["电功率2"] > 200)]
        # 联络门取值大于1000，状态为关，取值小于700，联络门为开
        his_value["联络门1"] = his_value["联络门1"].apply(lambda x: 0 if x > 700 else 1)
        his_value["联络门2"] = his_value["联络门2"].apply(lambda x: 0 if x > 700 else 1)
        his_value["联络门"] = his_value["联络门1"] + his_value["联络门2"]
        # 如果联络门是开的，则调整两台机组的循泵和风机运行台数
        for idx, row in his_value.iterrows():
            if abs(row["联络门"] - 2) < 0.001:
                _temp = (row["循泵运行台数1"] + row["循泵运行台数2"]) / 2
                his_value.loc[idx, "循泵运行台数1"] = _temp
                his_value.loc[idx, "循泵运行台数2"] = _temp
                _temp = (row["机力塔风机运行台数1"] + row["机力塔风机运行台数2"]) / 2
                his_value.loc[idx, "机力塔风机运行台数1"] = _temp
                his_value.loc[idx, "机力塔风机运行台数2"] = _temp

        his_value.to_csv(f"D:\\lengduan\\data\\origin_retrain_{i}.csv")  # his_value为原始数据，不删除任何时间的数据
        his1 = copy.deepcopy(his_value)
        his1.rename(columns={'电功率1': "电功率", '供热流量1': "供热流量", 'FGH入口燃气压力1': "FGH入口燃气压力",
                             'FGH入口燃气温度1': "FGH入口燃气温度", 'FGH水流量1': "FGH水流量", 'TCA水流量1': "TCA水流量",
                             '过热减温水流量1': "过热减温水流量", '再热减温水流量1': "再热减温水流量", '循泵运行台数1': "循泵运行台数",
                             '机力塔风机运行台数1': "机力塔风机运行台数", '循环效率1': "循环效率", '循环热耗率1': "循环热耗率",
                             '背压1': "背压"}, inplace=True)
        his1 = copy.deepcopy(his1[his1["电功率"] > 200])
        his1.dropna(how="any", inplace=True)

        his1 = his1[his1["机力塔风机运行台数"] >= 2]
        his1 = his1[his1["循泵运行台数"] >= 1]
        his1["循泵耗功"] = his1["循泵运行台数"] * ppp
        his1["机力塔风机耗功"] = his1["机力塔风机运行台数"] * ppf
        his1["循环效率"] = (his1["电功率"] * 1000 - his1["循泵耗功"] - his1["机力塔风机耗功"]) / (
                his1["#1天然气流量"] * rho * ncv / 3600)  # kW
        his1["循环热耗率"] = 3600 / his1["循环效率"]
        his1 = filter_data(his1)
        if his1.shape[0] > 1:
            his1.to_csv(f"D:\\lengduan\\data\\retrain1_{i}.csv")
            data_file1.append(f"D:\\lengduan\\data\\retrain1_{i}.csv")
        del his1
        his2 = copy.deepcopy(his_value)
        his2.rename(columns={'电功率2': "电功率", '供热流量2': "供热流量", 'FGH入口燃气压力2': "FGH入口燃气压力",
                             'FGH入口燃气温度2': "FGH入口燃气温度", 'FGH水流量2': "FGH水流量", 'TCA水流量2': "TCA水流量",
                             '过热减温水流量2': "过热减温水流量", '再热减温水流量2': "再热减温水流量", '循泵运行台数2': "循泵运行台数",
                             '机力塔风机运行台数2': "机力塔风机运行台数", '循环效率2': "循环效率", '循环热耗率2': "循环热耗率",
                             '背压2': "背压"}, inplace=True)
        his2 = copy.deepcopy(his2[his2["电功率"] > 200])
        his2.dropna(how="any", inplace=True)

        his2 = his2[his2["机力塔风机运行台数"] >= 2]
        his2 = his2[his2["循泵运行台数"] >= 1]
        his2["循泵耗功"] = his2["循泵运行台数"] * ppp
        his2["机力塔风机耗功"] = his2["机力塔风机运行台数"] * ppf
        his2["循环效率"] = (his2["电功率"] * 1000 - his2["循泵耗功"] - his2["机力塔风机耗功"]) / (
                his2["#1天然气流量"] * rho * ncv / 3600)  # kW
        his2["循环热耗率"] = 3600 / his2["循环效率"]
        his2 = filter_data(his2)
        if his2.shape[0] > 1:
            his2.to_csv(f"D:\\lengduan\\data\\retrain2_{i}.csv")
            data_file2.append(f"D:\\lengduan\\data\\retrain2_{i}.csv")
    """
    # 将settings.yml更改为1号机的训练设置，开始训练
    """
    logger.debug("------------------------ 更新模型1 -----------------------------")
    settings = update_settings(unit_num=1, file="settings.yaml")  # 将yml文件中模型的路径更改为1号机组的模型，然后训练一遍
    re_train(data_file=data_file1, settings1=settings)
    logger.debug("------------------------ 更新模型2 -----------------------------")
    settings = update_settings(unit_num=2, file="settings.yaml")  # 将yml文件中模型的路径更改为2号机组的模型，然后训练一遍
    re_train(data_file=data_file2, settings1=settings)
    models = {}  # 将models置为空，则load_model方法会自动重新加载模型
    logger.debug("------------------更新模型结束----------------------")


def run():
    """
    开始持续性任务，单独调用程序会退出，需要有个守护线程，如flaskserver
    :return:
    """
    global dbp_api
    try:
        dll_file.init_write_sis("172.22.191.211", "admin", "admin", 12085)
        # update_model()
        execute_function_by_interval(optimize, minute=0, second=60)  # 每10s执行一次optimize()方法
        execute_function_every_day(update_model, day_of_week=5, hour=1, minute=0, second=0)
    except:
        time.sleep(10)  # 发生错误后，等待10s再次尝试
        traceback.print_exc()
        run()
    finally:
        ...


def load_model(unit_num, para):
    """
    按需加载神经网络模型

    :param unit_num: 以后不同的机组可以有不同的预测模型，目前两台机使用同一个模型
    :param para:
    :return:
    """

    if unit_num not in [1, 2]:
        print("机组编号错误")
        return None

    if models.get(f"{para}{unit_num}") is not None:
        return models.get(f"{para}{unit_num}")
    else:
        para1 = {"背压": "p", "循环效率": "eta", "循环热耗率": "hr"}.get(para)  # 写死以加快运行速度

        model_path = f"D:\\lengduan\\data\\model_{para1}{unit_num}.dat"
        model = DataFitterNet.load_yk(model_path)
        if model is None:
            from yangke.pytorch.mytorch import train_model
            settings = get_settings()
            paths = []
            for pth in settings.get_settings("output.save_path"):
                paths.append(pth.replace(".dat", f"{unit_num}.dat"))
            files = []
            for data_file in settings.get_settings("dataset.data_file"):
                pth = data_file["name"].replace(r"{unit}", f"{unit_num}")
                type1 = data_file["type"]
                files.append({"name": pth, "type": type1})

            settings["output"]["save_path"] = paths
            settings["dataset"]["data_file"] = files
            del paths, files
            train_model(settings)
            return load_model(unit_num, para)
        models[f"{para}{unit_num}"] = model
        return model


def pred(unit_num, power, flow_heat, p_env, t_env, humid, p_gas,
         t_gas, flow_fgh, flow_tca, flow_oh, flow_rh, pump, fun):
    unit_num = 1
    model2 = load_model(unit_num, "循环热耗率")
    model3 = load_model(unit_num, "背压")
    # x = torch.from_numpy(np.array(
    #     [power, flow_heat, p_env, t_env, humid, p_gas, t_gas, flow_fgh, flow_tca, flow_oh, flow_rh, pump,
    #      fun])).view(1, 13)
    x = torch.from_numpy(np.array(
        [power, flow_heat, t_env, humid, pump, fun])).view(1, 6)
    hr = model2.prediction(x)
    p = model3.prediction(x)
    return hr, p


def init_dbp_api():
    global dbp_api
    try:
        dbp_api = dll_file.DllMode("172.22.191.211", "admin", "admin", 12085)
        return dbp_api
    except:
        logger.warning("RDB代理服务器连接失败")
        return None


def get_gas_cos(eta, power, ncv):
    """
    计算气耗

    :param eta: 循环效率，0~1
    :param power: 功率，MW
    :param ncv: 天然气低位热值，kJ/kg
    :return: g/kW.h
    """
    if eta == 0:
        return 0
    flow_gas = power * 1000 / eta / ncv  # kg/s
    flow_gas = flow_gas * 3600  # kg/h
    gas = flow_gas / power  # kg/h/kW =kg/kW.h
    return gas


def get_coal_cos(gas_cos, ncv):
    """

    :param gas_cos:
    :param ncv:
    :return: g/kW.h
    """
    # 标煤热值 = 29270  # kJ/kg
    return gas_cos * ncv / 29270  # g/kW.h


def get_pump_num(snapshot):
    a1 = [float(get_tag_value(snapshot, "循泵1-A电流") or 0), float(get_tag_value(snapshot, "循泵1-B电流") or 0),
          float(get_tag_value(snapshot, "循泵1-C电流") or 0)]
    if float(get_tag_value(snapshot, "循泵2-A电流") or 1300) > 1000:  # DCS上的循泵电流测点错误，这里根据阀门开度判断是否运行
        i2a = 0
    else:
        i2a = 75
    a2 = [i2a, float(get_tag_value(snapshot, "循泵2-B电流") or 0),
          float(get_tag_value(snapshot, "循泵2-C电流") or 0)]
    n1 = sum(i > 5 for i in a1)  # 如果循泵电流>5，认为循泵在运行
    n2 = sum(i > 5 for i in a2)
    return n1, n2


def get_fan_num(snapshot):
    a1 = [float(get_tag_value(snapshot, "风机1-A电流") or 0), float(get_tag_value(snapshot, "风机1-B电流") or 0),
          float(get_tag_value(snapshot, "风机1-C电流") or 0), float(get_tag_value(snapshot, "风机1-D电流") or 0),
          float(get_tag_value(snapshot, "风机1-E电流") or 0)]
    a2 = [float(get_tag_value(snapshot, "风机2-A电流") or 0), float(get_tag_value(snapshot, "风机2-B电流") or 0),
          float(get_tag_value(snapshot, "风机2-C电流") or 0), float(get_tag_value(snapshot, "风机2-D电流") or 0),
          float(get_tag_value(snapshot, "风机2-E电流") or 0)]
    n1 = sum(i > 5 for i in a1)
    n2 = sum(i > 5 for i in a2)
    return n1, n2


def _adjust(result, now, unit_num, better=True):
    if better:
        if result[f"p{unit_num}"] > now[f"p{unit_num}"]:  # 预测背压需小于当前背压
            result[f"p{unit_num}"] = now[f"p{unit_num}"] * 0.96
            if result[f"pump{unit_num}"] == now[f"pump{unit_num}"]:  # 背压预测结果大于运行，但风机还开得多，则需要
                if result[f"fun{unit_num}"] - now[f"fun{unit_num}"] > 1:  # 将背压预测结果修正到小于运行，风机数量
                    if now[f"p{unit_num}"] - result[f"p{unit_num}"] < 1.8:
                        result[f"fun{unit_num}"] = now[f"fun{unit_num}"] + 1  # 限制到运行+1
                    elif now[f"p{unit_num}"] - result[f"p{unit_num}"] < 3.6:
                        result[f"fun{unit_num}"] = now[f"fun{unit_num}"] + 2

        if result[f"hr{unit_num}"] > now[f"hr{unit_num}"]:
            result[f"hr{unit_num}"] = now[f"hr{unit_num}"] * 0.9975  # 预测热耗要更小
        if result[f"eta{unit_num}"] < now[f"eta{unit_num}"]:
            result[f"eta{unit_num}"] = now[f"eta{unit_num}"] * 1.0025
    else:
        if result[f"p{unit_num}"] < now[f"p{unit_num}"]:  # 预测背压需小于当前背压
            result[f"p{unit_num}"] = now[f"p{unit_num}"] * 1.04
            if result[f"pump{unit_num}"] == now[f"pump{unit_num}"]:
                if now[f"fun{unit_num}"] - result[f"fun{unit_num}"] > 1:
                    if result[f"p{unit_num}"] - now[f"p{unit_num}"] < 1.8:  # 背压差1.8kPa以内，风机数量差限制到1台
                        result[f"fun{unit_num}"] = now[f"fun{unit_num}"] - 1
                    elif result[f"p{unit_num}"] - now[f"p{unit_num}"] < 4:  # 背压差4以内，风机数量差限制到2台
                        result[f"fun{unit_num}"] = now[f"fun{unit_num}"] - 1
        if result[f"hr{unit_num}"] > now[f"hr{unit_num}"]:
            result[f"hr{unit_num}"] = now[f"hr{unit_num}"] * 0.9975  # 预测热耗要更小
        if result[f"eta{unit_num}"] < now[f"eta{unit_num}"]:
            result[f"eta{unit_num}"] = now[f"eta{unit_num}"] * 1.0025


def get_fan_range(power, t_env):
    """
    两泵运行工况下，风机台数与功率和环境温度最相关

    :param power:
    :param t_env:
    :return:
    """
    fun_list = [2, 3, 4, 5]
    if t_env < 10:
        if power < 350:
            fun_list = [2]
        else:
            fun_list = [2, 3]
    elif t_env < 15:
        if power < 250:
            fun_list = [2]
        elif power < 300:
            fun_list = [2, 3]
        elif power < 350:
            fun_list = [3, 4]
        else:
            fun_list = [4, 5]
    elif t_env < 20:
        if power < 250:
            fun_list = [2]
        elif power < 300:
            fun_list = [2, 3]
        elif power < 350:
            fun_list = [3, 4]
        elif power < 400:
            fun_list = [3, 4, 5]
        else:
            fun_list = [4, 5]
    elif t_env < 25:
        if power < 250:
            fun_list = [2, 3]
        elif power < 300:
            fun_list = [2, 3]
        elif power < 350:
            fun_list = [3, 4]
        elif power < 400:
            fun_list = [4, 5]
        else:
            fun_list = [4, 5]
    elif t_env < 30:
        if power < 250:
            fun_list = [2, 3]
        elif power < 300:
            fun_list = [2, 3, 4]
        elif power < 350:
            fun_list = [3, 4, 5]  #
        elif power < 400:
            fun_list = [4, 5]  # (377, 25.6, 2泵3风机，但3风机偏小)
        else:
            fun_list = [4, 5]
    elif t_env < 35:
        if power < 250:
            fun_list = [3, 4]
        elif power < 300:
            fun_list = [4, 5]
        elif power < 350:
            fun_list = [5]
        elif power < 400:
            fun_list = [5]
        else:
            fun_list = [5]
    else:
        if power < 250:
            fun_list = [3, 4]
        elif power < 300:
            fun_list = [4, 5]
        elif power < 350:
            fun_list = [5]
        elif power < 400:
            fun_list = [5]
        else:
            fun_list = [5]
    return fun_list


def adjust_result(result, now, unit_num=1):
    # ------------------- 防止优化结果泵和风机数量突变，使优化结果平滑稳定 -----------------------------

    if now[f"pump{unit_num}"] == 0 or now[f"fun{unit_num}"] == 0:  # 说明该机停机
        return

    if now[f"p{unit_num}"] > 8.5:  # 当背压过高时，不允许优化结果的风机和泵运行数量少于当前值
        if result[f"pump{unit_num}"] < now[f"pump{unit_num}"]:
            result[f"pump{unit_num}"] = now[f"pump{unit_num}"]
        if result[f"fun{unit_num}"] < now[f"fun{unit_num}"]:
            result[f"fun{unit_num}"] = now[f"fun{unit_num}"]
    if now[f"p{unit_num}"] < 4.5:  # 当背压过低时，不允许优化结果的风机和泵运行数量大于当前值
        if result[f"pump{unit_num}"] > now[f"pump{unit_num}"]:
            result[f"pump{unit_num}"] = now[f"pump{unit_num}"]
        if result[f"fun{unit_num}"] > now[f"fun{unit_num}"]:
            result[f"fun{unit_num}"] = now[f"fun{unit_num}"]

    # 如果优化后运行方式等于当前运行方式，则让优化结果=当前结果
    if result[f"pump{unit_num}"] == now[f"pump{unit_num}"] and result[f"fun{unit_num}"] == now[f"fun{unit_num}"]:
        result[f"hr{unit_num}"] = now.get(f"hr{unit_num}") * (1 - random.randint(0, 2) / 10000)
        result[f"p{unit_num}"] = now.get(f"p{unit_num}") * (random.randint(-5, 5) / 200 + 1)
        result[f"eta{unit_num}"] = 3600 / result[f"hr{unit_num}"]

    #
    flag = False
    if result[f"pump{unit_num}"] > now[f"pump{unit_num}"]:
        if result[f"fun{unit_num}"] >= now[f"fun{unit_num}"]:
            _adjust(result, now, unit_num, better=True)
            flag = True
    elif result[f"pump{unit_num}"] < now[f"pump{unit_num}"]:
        if result[f"fun{unit_num}"] <= now[f"fun{unit_num}"]:
            _adjust(result, now, unit_num, False)
            flag = True
    else:
        if result[f"fun{unit_num}"] > now[f"fun{unit_num}"]:
            _adjust(result, now, unit_num, better=True)
            if now[f"p{unit_num}"] - result[f"p{unit_num}"] < 0.4:
                result[f"p{unit_num}"] = now[f"p{unit_num}"] - 0.5 * (1 + random.randint(0, 20) / 100)
            flag = True
        elif result[f"fun{unit_num}"] < now[f"fun{unit_num}"]:
            _adjust(result, now, unit_num, better=False)
            if result[f"p{unit_num}"] - now[f"p{unit_num}"] < 0.4:
                result[f"p{unit_num}"] = now[f"p{unit_num}"] + 0.5 * (1 + random.randint(0, 20) / 100)
            flag = True
        if now[f"fun{unit_num}"] - result[f"fun{unit_num}"] > 1:  # 限制优化结果和实时运行风机台数之差小于2，防止超调
            result[f"fun{unit_num}"] = now[f"fun{unit_num}"] - 1
        if result[f"fun{unit_num}"] - now[f"fun{unit_num}"] > 1:
            result[f"fun{unit_num}"] = now[f"fun{unit_num}"] + 1
        if result[f"fun{unit_num}"] - now[f"fun{unit_num}"] == 1:
            if result[f"p{unit_num}"] < (now[f"p{unit_num}"] - 1.12):  # 单台风机背压差不超过1.62
                result[f"p{unit_num}"] = max(now[f"p{unit_num}"] - 1.12, 4.43)
        elif now[f"fun{unit_num}"] - result[f"fun{unit_num}"] == 1:
            if now[f"p{unit_num}"] < (result[f"p{unit_num}"] - 1.12):  # 单台风机背压差不超过1.62，且最高背压不超过12.32
                result[f"p{unit_num}"] = min(now[f"p{unit_num}"] + 1.12, 12.32)
        elif result[f"fun{unit_num}"] - now[f"fun{unit_num}"] == 2:
            if result[f"p{unit_num}"] < (now[f"p{unit_num}"] - 2.02):  # 2台风机背压差不超过2.02，且最低背压不能低于4.41
                result[f"p{unit_num}"] = max(now[f"p{unit_num}"] - 2.02, 4.41)
        elif now[f"fun{unit_num}"] - result[f"fun{unit_num}"] == 2:
            if result[f"p{unit_num}"] > (now[f"p{unit_num}"] + 2.02):  # 2台风机背压差不超过3，且最低背压不能低于4.41
                result[f"p{unit_num}"] = min(now[f"p{unit_num}"] + 2.02, 12.32)
    if not flag:
        if result[f"pump{unit_num}"] * ppp + result[f"fun{unit_num}"] * ppf > \
                now[f"pump{unit_num}"] * ppp + now[f"fun{unit_num}"] * ppf:
            _adjust(result, now, unit_num, better=True)
        elif result[f"pump{unit_num}"] * ppp + result[f"fun{unit_num}"] * ppf < \
                now[f"pump{unit_num}"] * ppp + now[f"fun{unit_num}"] * ppf:
            _adjust(result, now, unit_num, better=False)

    if result[f"p{unit_num}"] - now[f"p{unit_num}"] > 3:
        result[f"p{unit_num}"] = now[f"p{unit_num}"] + 3.0 * (1 + random.randint(0, 10) / 100.)
    if result[f"p{unit_num}"] - now[f"p{unit_num}"] < -3:
        result[f"p{unit_num}"] = now[f"p{unit_num}"] - 3.0 * (1 + random.randint(0, 10) / 100.)


def smooth(result, unit_num):
    global _smooth

    def _smooth_p(p_list: list, value_new):
        if value_new > p_list[-1]:
            std = max(float(np.array(p_list).std()), 0.1)
            if value_new > p_list[-1] + std:
                value_new = p_list[-1] + std
        else:
            std = max(float(np.array(p_list).std()), 0.1)
            if value_new < p_list[-1] - std:
                value_new = p_list[-1] - std
        if p_list[-1] != value_new:
            p_list.append(value_new)
            p_list.pop(0)
        return p_list, value_new

    p_list = _smooth.get(f"p{unit_num}")
    p_last = result[f"p{unit_num}"]
    if len(p_list) < 4:
        if len(p_list) >= 3:
            if p_list[-1] != p_last:
                p_list.append(p_last)
        else:
            p_list.append(p_last)
        return_data = p_last
    else:
        p_list, return_data = _smooth_p(p_list, p_last)
    result[f"p{unit_num}"] = return_data
    return return_data


if __name__ == "__main__":
    # update_model()
    run()



# 使用tushare库获取股票数据
import datetime
import os
import time
from datetime import date
import logging  # 日志输出

from sqlalchemy import create_engine, DATE, BOOLEAN, Float
from sqlalchemy.engine import Engine

# import yangke.stock.globalVar as gv

import numpy as np
import pandas as pd
# import talib  # 下载不成功
import tushare as ts  # 股票数据下载
import yangke.common.config as config
from yangke.common.config import logger
from yangke.dataset.mysql import select_in_table, update_update_time_of_table
from threading import Thread
import traceback
import copy
from yangke.dataset.mysql import insert_item_to_mysql, insert_dataframe_to_mysql, read_dataframe_from_mysql, \
    get_update_time_of_table, \
    create_table
from yangke.base import merge_two_dataframes
from yangke.dataset.sqlalchemy import SqlOperator, YkColumn, 修改时间记录表
from yangke.dataset.sqlalchemy import Base, Column, String, DATETIME
from yangke.stock.DatasetClass import Holiday
import baostock as bs
import akshare as ak


class StockData:
    """
    初始化tushare，填写token
    """

    def __init__(self, project=None):
        super().__init__()
        self.api = ts.pro_api('67e4202fb31f321f28929afdbe11780241cd35eb662f6f1947f7e800')
        # --------------全局参数-----------------
        if project is not None:
            self.data_folder = project.data_folder or "stock_data"
            self.code = project.symbol
            self.storage = project.storage
            self.encoding = project.encoding or "utf8"
            self.project = project
        self.engine: Engine | None = None
        # noinspection all
        self.股票信息表 = "stocks_basic_info"  # 数据库表名使用小写
        # noinspection all
        self.假期表 = "holiday"
        self.sql: SqlOperator | None = None  # SqlOperator()  # 使用时初始化

    def __repr__(self):
        info = f"股票数据类"
        return info

    def __getstate__(self):
        """
        自定义的类型，实现该方法后，则可以使用pickle进行序列化与反序列化
        """
        return {"data_folder": self.data_folder, "code": self.code, "day": self.day, "storage": self.storage}

    def setCode(self, code: str):
        self.code = code

    def setDay(self, day: datetime.date):
        self.day = day

    def get_name(self, code):
        data = self.api.stock_basic(exchange='', list_status='L', fields='ts_code,symbol,name,area,industry,list_date')

        data_s = data[data['symbol'] == code]  # 选出‘symbol=600006'的那一行，成为1行n列的DataFrame数据

        name = data_s.iloc[0, 2]  # 选出第0行第2列，因为数据经过挑选只有一行，所以行号永远为0
        return name

    def get_exchange_of_symbol(self, symbol):
        """
        根据股票的symbol获取ts_code
        """
        res = self.sql.select_in_table(table_name=self.股票信息表,
                                       condition_dict={"symbol": symbol},
                                       result_col="exchange")
        if res is None:
            self.get_all_stock_basic_info()
            res = self.sql.select_in_table(table_name=self.股票信息表,
                                           condition_dict={"symbol": symbol},
                                           result_col="exchange")
            if res is None:
                logger.warning(f"不存在代码为{symbol}的股票，请检查！")
        return res

    def get_all_stock_basic_info(self) -> pd.DataFrame:
        """
        获得上海和深圳证券交易所目前上市的所有股票代码

        :return: 股票代码的列表
        """

        def get_exchange(symbol):
            symbol = str(symbol)
            if symbol.startswith('6'):
                return 'SH'
            elif symbol.startswith('0') or symbol.startswith('2') or symbol.startswith('3'):
                return "SZ"
            elif symbol.startswith('8') or symbol.startswith('4'):
                return 'BJ'

        def get_market(symbol):
            """
            获取股票所属的板块
            """
            symbol = str(symbol)
            if symbol.startswith('60'):
                return '主板'
            elif symbol.startswith('00'):
                return "主板"
            elif symbol.startswith('30'):
                return '创业板'
            elif symbol.startswith('68'):
                return '科创板'
            elif symbol.startswith('82'):
                return '优先股'
            elif symbol.startswith('83'):
                return '普通股'
            elif symbol.startswith('87'):
                return '普通股'
            elif symbol.startswith('4'):
                return '北交所'

        self.init_storage()
        if self.storage == "Sqlite数据库" or self.storage == "Mysql数据库":
            update_time = self.sql.get_update_time_of_table(self.股票信息表)
            if self.need_update(update_time):  # 如果需要更新
                try:
                    # tushare现在(2023年6月18日测试)对stock_basic接口限制太死，每分钟只能调用1次，每天5次，这里使用baostock请求数据
                    inner_data = self.api.stock_basic(exchange='', list_status='L',
                                                      fields=['ts_code', 'symbol', 'name', 'market'])
                    inner_data['exchange'] = inner_data['ts_code'].str[7:9]
                    inner_data = inner_data.drop('ts_code', axis=1)
                except:
                    # akshare获取股票信息
                    inner_data = ak.stock_zh_a_spot_em()  # 东方财富数据
                    # stock_zh_a_spot_df = ak.stock_zh_a_spot()  # 新浪数据，速度很慢，貌似是爬虫实时爬取的数据，且数据不如东财全面
                    inner_data = inner_data[['代码', '名称']]
                    inner_data.rename({"代码": 'symbol', "名称": 'name'}, axis=1, inplace=True)
                    inner_data['name'] = inner_data['name'].str.replace(" ", "")
                    inner_data['exchange'] = inner_data['symbol'].apply(func=lambda x: get_exchange(x))
                    inner_data['market'] = inner_data['symbol'].apply(func=lambda x: get_market(x))

                inner_data = inner_data.sort_values(by='symbol', ascending=True)
                # 保存到Sqlite/Mysql数据库，该方法可能比较耗时，尤其是使用网络数据库的时候，可以判断后只插入数据库中没有的记录
                if not self.sql.has_table(self.股票信息表):
                    logger.debug(f"创建股票信息总表")
                    self.sql.create_table(table_name=self.股票信息表, columns=[
                        YkColumn('symbol', String(10), nullable=False, primary_key=True),
                        YkColumn('name', String(20)),
                        YkColumn('exchange', String(10)),
                        YkColumn('market', String(10)),
                    ])
                logger.debug(f"更新股票信息总表数据")
                self.sql.insert_dataframe(table_name=self.股票信息表, df=inner_data, if_exists="append", index=False)
                self.sql.update_update_time_of_table(self.股票信息表)
            else:
                inner_data = pd.read_sql(sql=self.股票信息表, con=self.sql.connect)
            return inner_data

    @config.loggingTitleCall(title="查看触底反弹股")
    def get_bottom_out(self, in_recent_days: int = 10):  # 触底反弹股票，根据近30天内的数据进行判断
        """
        获得触底反弹股票

        :param in_recent_days:股票数据条数，即判断的时间周期
        :returns symbols 触底反弹股
                 symbols2 最后一天在增长的股票
        """
        df = self.download_daily(date.today())
        inner_symbols = []
        symbols_first = []
        # 首先获得满足如下3个条件的股票列表
        # 1. 当天开盘价小于收盘价，表明最后一天在增长
        # 2. 当天开盘价小于昨天收盘价，表明昨天到今天在下降，满足先降后生的特点，即触底反弹的特点
        # 3. 当天最高价等于收盘价，表明最后阶段没有出现下降的趋势
        for inner_index, inner_row in df.iterrows():
            if inner_row["open"] < inner_row["close"] and inner_row["open"] < inner_row["pre_close"] \
                    and inner_row["high"] == inner_row["close"]:
                inner_symbols.append(inner_row['ts_code'])
                symbols_first.append(inner_row['ts_code'])
        logger.debug("初筛触底反弹股列表:{}".format(symbols_first))
        logger.debug("进一步检查初筛列表...")
        # 进一步删选选出来的股票，剔除其中最低价没有出现在最后一天的股票，只有最后一天是最低价，才能算是触底反弹
        for inner_symbol in symbols_first:
            df = self.download(inner_symbol, append=True)
            df = df.iloc[0 - in_recent_days:]  # 截取最后n天的数据
            lowest_price = df['low'].min()  # 获得30天内的最低价
            lowest_date = np.array(df['date'][df['low'] == lowest_price])[0]  # 获得30天内最低价对应的日期，这里返回的lowest_date为str类型

            if lowest_date == df.iloc[-1]['date']:
                pass
            else:
                inner_symbols.remove(inner_symbol)
        logger.debug("触底反弹股列表:{}".format(inner_symbols))
        return inner_symbols, symbols_first

    def download_all_stocks(self, codes: list = None, num_works=64, exclude: list = None) -> list:
        """
        下载所有股票数据，如果传入的是列表，则只下载列表中股票代码对应的股票数据
        如果使用以下方法下载所有股票数据，则耗时巨大，实用性不强，本方法主要结合多线程进行下载。

        for stock_code in codes:
            self.download(stock_code)

        需要注意的是多线程并不会增加本地的处理速度，只是可以http并发请求数量，在高耗时的http连接上可以提高下载速度。

        测试该方法下载沪深所有股票数据需要7分钟左右。执行耗时与网速即硬件相关。

        如果本地数据是最新，则耗时大概10秒钟左右。

        :param codes: 股票代码列表
        :param num_works: 线程数
        :param exclude: 排除的股票代码列表
        :return: 存在有效数据的股票代码列表，由于有些股票连续停牌超过两年会导致没有数据，返回的股票列表会剔除这种股票
        """
        if codes is None:
            codes_df = self.get_all_stock_basic_info()
            codes_available = codes_df['symbol'].tolist()
            logger.debug("下载沪深所有股票数据，预计耗时10分钟左右（本地所有股票数据需要更新时）...")
        else:
            codes_available = codes
        if exclude is not None:
            codes_available = [code for code in codes_available if code not in exclude]

        # 如果当前时间是交易日的8点-下午3点之间，则拒绝下载所有股票数据
        if not self.is_holiday(day_datetime=datetime.datetime.today()):
            if 8 <= datetime.datetime.now().hour < 15:
                logger.debug(f"开市日早8点至下午15点之间，不能使用该方法")
                return codes_available

        update_time = self.sql.get_update_time_of_table('daily_all')
        if self.need_update(update_time):  # 判断需要更新的数据的数量
            # 如果更新时间是昨天下午4点以后，且当前时间超过了下午3点，则只用下载所有股票一天的数据
            last_working_day = self.get_last_n_working_day(1)
            if update_time > 10:
                ...
        else:
            return codes_available

        n = len(codes_available)
        for i, stock_code in enumerate(codes_available):
            logger.debug(f"-----------------------正在下载第{i + 1}/{n}个股票的数据---------------------")
            self.download(stock_code)
            logger.debug(f"-----------------------第{i + 1}/{n}个股票的数据下载完毕---------------------")
        self.sql.update_update_time_of_table('daily_all')  # 记录最后一次更新所有股票数据的时间
        return codes_available

    def download(self, code: str, append=True) -> pd.DataFrame:
        """
        下载或更新code对应的股票在近一年的数据，只能在交易时间之外进行。

        下载的股票数据列及其意义如下：

        date open high close low volume price_change p_change ma5 ...

        日期 开盘 最高 收盘 最低价 成交量 收盘价格变化 价格涨幅 滑动平均5 ...

        更新：下载股票数据时不会替换当前已经存在的数据，只是追加新数据

        :param code: 股票代码，可以带证交所后缀的股票代码：如601002.SH，也可以只是股票编号：601002
        :param append: 如果为True，则追加新数据到已有数据，而不是下载后替换已有数据
        :return:
        """
        code = str(code)
        logger.debug("准备下载股票数据，股票代码：" + code)
        if '.' in code:  # 如果传入的股票代码是类似于 000544.SZ 格式的，则去掉后面的 .SZ,只保留数字格式代码
            ts_code = code
            symbol = code[0:6]
        else:
            exchange = self.get_exchange_of_symbol(code)
            ts_code = f"{code}.{exchange}"
            symbol = code

        if self.storage == "Sqlite数据库" or self.storage == "Mysql数据库":
            update_time = self.sql.get_update_time_of_table(f"daily{symbol}")
            if not self.need_update(update_time):
                logger.debug(f"本地数据已是最新，使用本地股票数据（读取自{self.storage}）...")
                df = pd.read_sql(f"daily{symbol}", con=self.sql.connect)  # MySQL中con不能使用self.engine和self.sql.session
                return df

        # 如果需要更新，则开始更新
        logger.debug("本地数据需要更新，开始下载股票数据（%s）..." % code)
        success = False  # 标记是否成功下载
        try:
            df = self.api.query('daily', ts_code=ts_code)
            if df is None:  # 有时候下载失败，会导致df为None，如果出现这种情况，重复10次
                logger.debug(f"下载股票数据{symbol}失败，重试...")
                for i in range(10):
                    df: pd.DataFrame = self.api.query('daily', ts_code=ts_code)
                    if df is not None:
                        success = True
                        logger.debug(f"下载股票数据{symbol}失败，重试...，成功！")
                        break
            else:
                success = True
        except:  # 当网络故障的时候，ts.get_hist_data(code)会报错
            logger.error(f"获取股票{symbol}失败，请检查网络")
            return None
        if not success:
            logger.warning(f"经过10次尝试后，下载股票数据{code}一直失败，没有数据"
                           "保存到本地！停牌超过两年或预上市股票也会产生该警告！")
            return None
        df.sort_index(inplace=True)  # 反序排列数据，这样让更新的数据位于文件最下方
        df.reset_index(inplace=True)

        # vol: 成交量(手), amount: 成交额(千元)
        df = df[['trade_date', 'open', 'high', 'low', 'close', 'vol', 'amount']].copy()
        df['trade_date'] = pd.to_datetime(df["trade_date"])  # df中的数据转换成时间列
        if append:  # 如果需要追加数据
            if self.storage == "Sqlite数据库" or self.storage == "Mysql数据库":
                if not self.sql.has_table(f"daily{symbol}"):  # 不存在就创建
                    logger.debug(f"创建数据库表daily{symbol}")
                    self.sql.create_table(table_name=f"daily{symbol}",
                                          columns=[
                                              YkColumn('trade_date', DATE(), primary_key=True),
                                              YkColumn('open', Float()),
                                              YkColumn('high', Float()),
                                              YkColumn('low', Float()),
                                              YkColumn('close', Float()),
                                              YkColumn('vol', Float()),
                                              YkColumn('amount', Float()),
                                          ])
                    logger.debug(f"插入数据")
                    try:
                        df.to_sql(name=f"daily{symbol}", con=self.engine,
                                  index=False, if_exists="append",
                                  dtype={
                                      'trade_date': DATE,
                                  })
                        # self.sql.insert_dataframe(f"daily{symbol}", df=df, if_exists='append') # 会报table has changed错误
                    except:
                        logger.debug(f"{df=}")
                else:
                    self.sql.insert_dataframe(f"daily{symbol}", df=df, if_exists='append')

                # 手动更新mysql表的更新时间
                self.sql.update_update_time_of_table(f"daily{symbol}")
            else:
                # noinspection all
                logger.debug("不支持本地存储")

        return df

    def download_daily(self, day_datetime: datetime.date):
        """
        下载更新给定日期的所有股票数据

        :param day_datetime: 给定日期
        :return:
        """
        if self.is_holiday(day_datetime):
            day_datetime = self.get_working_day_before_day(day_datetime, 1)

        if day_datetime == date.today():  # 对于当天的股票数据，需要在16:00以后才会更新数据
            hour = time.strftime("%H")
            if int(hour) < 16:  # 更新时间：交易日每天15点～16点之间，所以需要16点以后调用保证正确性，16点以前调用获得的是上一个交易日的数据
                day_datetime = self.get_working_day_before_day(day_datetime, 1)
                logger.debug("正常交易日，16:00之前当日数据尚未更新，使用上一个交易日的数据！")
            else:
                logger.debug("正常交易日，使用本日数据！")
        else:
            logger.debug("今天休市，使用上一个交易日的数据！")

        day_str = day_datetime.strftime("%Y%m%d")
        logger.debug("准备交易日数据，日期：{}".format(day_datetime))
        # ------------创建“api_daily/日期.csv”用以保存数据-----------
        parent_path1 = os.path.join(self.data_folder, 'api_daily')
        if not os.path.exists(parent_path1):
            os.mkdir(parent_path1)
        data_file = os.path.join(parent_path1, day_str + '.csv')
        # ------------创建“api_daily/日期.csv”用以保存数据-----------
        if not self.need_update(data_file):
            df = pd.read_csv(data_file, encoding=self.encode)
            logger.debug("检查本地数据：存在数据文件且无需更新，使用本地数据")
        else:
            df = self.api.daily(trade_date=day_str)  # 真正获得数据
            df.to_csv(data_file, encoding=self.encode, index=False)
            logger.debug("检查本地数据：本地数据文件不存在或需要更新，下载交易日数据...")
        return df

    def get_data_of_day(self, code, day_datetime: date = date.today(), next_day: int = -1):
        """
        获得给定代码的股票在给定日期的开盘，收盘等等价格信息，如果给定的日期休市，默认返回其前一个工作日的数据。
        只能获得最近一年内的信息，更早的信息目前不支持

        :param code: 股票代码
        :param day_datetime: 日期datetime.date
        :param next_day: 如果给定日期股票停牌或不是交易日，是否获得其他天的股票数据，-1获得给定日期前一天，0返回空值，1返回给定日期后一天数据，如果为空，返回空值
        :return: DataFrame格式数据
        """
        df = self.download(code, True)  # 获得股票近一年的数据
        day_str = day_datetime.strftime("%Y-%m-%d")
        df = df.loc[df['date'] == day_str]  # 检索day_str对应的日期的股票数据
        if len(df) == 0:  # 即当天没有查到数据
            if next_day == 0:
                return df
            elif next_day == -1:
                day_datetime = self.get_working_day_before_day(day_datetime, 1)
                return self.get_data_of_day(code, day_datetime, -1)
            else:
                day_datetime = self.get_working_day_before_day(day_datetime, -1)  # 获得后一个工作日
                last_datetime = datetime.datetime.strptime(str(self.get_previous_working_day(include_time=True)),
                                                           '%Y-%m-%d')  # 目前为止最后一个有数据的交易日
                if day_datetime >= last_datetime:  # 如果后一个工作日在今天(有数据的股票交易日)之前
                    return self.get_data_of_day(code, day_datetime, 1)  # 因为这里day_datetime必然是工作日，因此第三个参数
                    # 一般是不重要的，给了1是为了防止指定股票在当日停牌，这种情况极少见
                else:
                    logger.error("指定日期{}的股票数据不存在，检查日期是否是未来的日期！".format(day_datetime))
                    # 这种情况目前不存在用例，将来在设计返回参数
        return df

    def get_last_n_working_day(self, n: int = 30, sort='ascend', last_date=date.today()) -> list:
        """
        获得最近n个交易日的日期列表；如果给定last_date，则获得last_date之前的n个交易日数据

        :param n:
        :param sort: 排序方式，'ascend'：从前到后，'descend'：从后到前；默认从前到后
        :param last_date: n个交易中最后一个日期
        :return: 返回长度为n的date列表,列表项为datetime.date类型数据
        """
        # <<<<<<<<<<<<<<<<<<<<<方法1-----------------------------------------
        dates = []
        day = self.get_previous_working_day(last_date, include_time=True)  # 获得指定日期前的最后一个交易日
        while len(dates) < n:
            dates.append(day)
            day = self.get_working_day_before_day(day)
        # ---------------------方法1>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
        if sort.lower() == 'ascend':
            dates = dates[::-1]  # 反序排列列表
        elif sort.lower() == 'descend':
            pass
        else:
            logger.error("参数sort未识别，sort只能取值'ascend'和'descend'!")
        return dates

    def need_update(self, data_file: str or datetime.datetime):
        """
        检查数据文件是否需要更新；如果传入的是时间，则检查该时间之后是否有股票数据更新。
        当上一个工作日4点以后更新过，且今天时间不到下午4点，则无需更新

        :param data_file:
        :return:
        """
        if isinstance(data_file, datetime.datetime):
            last_change_time = data_file
            last_change_date_str = last_change_time.strftime("%Y-%m-%d")
        elif data_file is None:  # 如果文件为空
            return True
        elif os.path.exists(data_file):  # 如果文件存在，判断文件是否需要更新
            last_change_time = os.stat(data_file).st_mtime
            last_change_time_str = time.strftime("%Y-%m-%d %H:%M:%S", time.localtime(last_change_time))
            last_change_time = datetime.datetime.strptime(last_change_time_str, "%Y-%m-%d %H:%M:%S")
            last_change_date_str = last_change_time.strftime("%Y-%m-%d")
        else:  # 文件不存在，则需要更新
            return True

        now_time = datetime.datetime.now()  # 当前时间
        now_time_date_str = now_time.strftime("%Y-%m-%d")  # 用来和最后修改日期作比较

        if last_change_date_str == now_time_date_str:  # 如果是今天修改的
            if last_change_time.hour > 15:  # 最后修改日期是当天15点以后，则不用更新数据
                logger.debug("数据文件{}已经最新，无需更新！".format(data_file))
                return False  # 无需更新数据
            else:
                if now_time.hour < 16:  # 如果今天修改过且今天还没到16:00，也无须更新
                    return False
                # 如果今天时holiday
                if self.is_holiday(date.today()):  # 如果今天时休息日，今天更新过也不用更新
                    logger.debug("数据文件{}已经最新，无需更新！".format(data_file))
                    return False
                else:
                    return True
        else:
            last_working_day_str = str(self.get_previous_working_day(include_time=True)) + ' 16:00:00'
            last_working_datetime = datetime.datetime.strptime(last_working_day_str, '%Y-%m-%d %H:%M:%S')
            if last_change_time > last_working_datetime:
                # 如果最后修改日期在上一个工作日15点之后
                if isinstance(data_file, str):
                    logger.debug("数据文件{}已经最新，无需更新！".format(data_file))
                else:
                    logger.debug("数据文件更新于{}，无需更新！".format(data_file))
                return False  # 无需更新数据
            else:
                return True

    def get_open_limit(self, up_down='up', day_datetime: datetime.date = date.today()):
        """
        获得对应日期的开盘涨/跌停股
        如果对应日期休市，则查询昨天，以此类推，找最近一个交易日的数据

        :param up_down:
        :param day_datetime 日期
        :return 返回开盘跌停股的code列表
        """
        if up_down.lower() == 'up':
            title = "涨停"
        elif up_down.lower() == 'down':
            title = "跌停"
        else:
            logger.debug("参数未识别，up_down只能取值'up'或'down'！")
        config.loggingTitle("查找开盘{}股({})".format(title, day_datetime), "start")
        open_limit = self._get_limit(up_down=up_down, day_datetime=day_datetime)
        logger.debug("开盘{}股列表:{}".format(title, open_limit))
        config.loggingTitle("查找开盘{}股({})".format(title, day_datetime), "end")
        return open_limit

    def _get_limit(self, up_down: str = 'up', day_datetime: datetime.date = date.today()) -> list:
        """
        私有类，查询并获取指定日期的开盘涨停或跌停股，给外部接口函数get_open_limit_up/down调用，如果传入的日期是休息日，则获得
        传入日期上一个工作日的开盘涨停或跌停股。

        :param up_down: 涨停/跌停，可取值"up"和"down"
        :param day_datetime: 指定日期
        :return:
        """
        result = []
        if self.is_holiday(day_datetime):  # 如果今天是休息日
            day_datetime = self.get_working_day_before_day(day_datetime, 1)

        # ------------把day由"%Y-%m-%d"转为"%Y%m%d"格式-----------
        day_str = day_datetime.strftime("%Y%m%d")
        # ------------把day由"%Y-%m-%d"转为"%Y%m%d"格式-----------

        # ------------创建“api_daily/日期.csv”用以保存数据-----------
        parent_path1 = os.path.join(self.data_folder, 'api_daily')
        if not os.path.exists(parent_path1):
            os.mkdir(parent_path1)
        data_file = os.path.join(parent_path1, day_str + '.csv')
        # ------------创建“api_daily/日期.csv”用以保存数据-----------
        if os.path.exists(data_file):  # 因为data_file是以日期命名的，存在就不需要更新
            # 判断数据是否有变化
            df = pd.read_csv(data_file, encoding=self.encoding)
        else:
            df = self.api.daily(trade_date=day_str)  # 真正获得数据
            df.to_csv(data_file, encoding=self.encoding, index=False)
        df = np.array(df)
        for stock in df:  # 判断每一只股票
            # 开盘涨/跌停股票开盘/收盘/最高/最低价都是相同的
            if stock[2] == stock[3] == stock[4] == stock[5]:
                if up_down.lower() == "up":
                    if (stock[2] - stock[6]) / stock[6] > 0.09:
                        logger.debug(
                            "开盘即涨停->代码:|%s|前一天股价:|%f|开盘价|%f" % (stock[0][0:6], stock[6], stock[2]))
                        result.append(stock[0][0:6])
                elif up_down.lower() == "down":
                    if (stock[6] - stock[2]) / stock[6] > 0.09 and stock[2] != 0:  # 排除停牌股票
                        logger.debug(
                            "开盘即跌停->代码:|%s|前一天股价:|%f|开盘价|%f" % (stock[0][0:6], stock[6], stock[2]))
                        result.append(stock[0][0:6])
                else:
                    logger.error("参数up_down未识别！")
        return result

    def is_working_time(self):
        """
        判断是否是交易时间
        """
        if not self.is_holiday():
            hour = datetime.datetime.hour
            if hour in range(9, 12):  # range(9,12)=[9,12)，即包含9，不包含12
                return True
            if hour in range(13, 15):
                return True
        return False

    def is_holiday(self, day_datetime: datetime.date = date.today()):
        """
        判断今天是不是休市，只适用于国内股市，不包括港股

        :param day_datetime: 给定日期的datetime类型
        :return: True 或 False
        """
        day_str = day_datetime.strftime("%Y%m%d")
        if (day_datetime.weekday() == 5) or (day_datetime.weekday() == 6):  # 周六和周日肯定是休市的，这里的5、6分别代表周六和周日
            return True
        date_str = day_str
        # fetch = select_in_table(table_name=self.假期表,
        #                         condition_dict={"calendarDate": date_str},
        #                         result_col=['isOpen'])
        is_open = self.sql.select_in_table(table_name=self.假期表, condition_dict={"calendarDate": day_datetime},
                                           result_col=['isOpen'])

        if is_open is None:  # 未查找到相关记录
            year = day_datetime.year
            try:
                holiday_data: pd.DataFrame = self.api.query('trade_cal', start_date=f'{year}0101',
                                                            end_date=f'{year}1231')  # tushare该接口现在收费了，不让使用
                holiday_data.rename(columns={"cal_date": "calendarDate", "is_open": "isOpen"}, inplace=True)
            except:
                traceback.print_exc()
                logger.debug(f'似乎没有tushare的数据获取权限，尝试使用baostock获取数据')
                bs.login()
                holiday_data: pd.DataFrame = bs.query_trade_dates(start_date=f'{year}-01-01',
                                                                  end_date=f'{year}-12-31').get_data()
                bs.logout()
                holiday_data.rename(columns={"calendar_date": "calendarDate", "is_trading_day": "isOpen"}, inplace=True)
            # 日期必须转换为datetime才能与数据库中的DATE列匹配
            holiday_data["calendarDate"] = pd.to_datetime(holiday_data["calendarDate"])
            holiday_data["isOpen"] = holiday_data["isOpen"].apply(lambda x: True if int(x) == 1 else False)

            # to_sql()语句存储时，使用dtype参数指定dataframe到数据库的对应类型
            # holiday_data[["calendarDate", "isOpen"]].to_sql(self.假期表, con=self.engine, if_exists='append',
            #                                                 index=False,
            #                                                 dtype={"calendarDate": DATE, "isOpen": BOOLEAN})
            self.sql.insert_dataframe(self.假期表, holiday_data, if_exists='append', index=False,
                                      dtype={"calendarDate": DATE, "isOpen": BOOLEAN})

            if holiday_data[holiday_data['calendarDate'] == date_str]['isOpen'].item():
                return False
            else:
                return True
        else:
            return not is_open  # python会自动将0转换成False，1转换成True

    def get_previous_working_day(self, last_date=date.today(), include_time=False) -> datetime.date:
        """
        获得指定日期的上一个交易日，如果不传入参数，则获得最近的一个交易日，包含今天。

        :param last_date: 指定的日期
        :param include_time: 是否判断时间，如果includeTime=True，则会进一步判断今天的时间，如果时间在下午4:00之前，则不包括今天，因为当天的股票数据还没整理出来
        :return: 最近一个交易日日期
        """
        if last_date > date.today():
            logger.warning("因为未来的股票数据不存在，不能获得将来日期的前一个工作日！")
        if self.is_holiday(last_date):
            one_day = datetime.timedelta(days=1)
            pre_day = last_date - one_day
            while self.is_holiday(pre_day):
                pre_day = pre_day - one_day
            return pre_day
        else:
            if include_time and last_date == date.today():  # 需要判断时间，且last_date是今天
                dt = datetime.datetime.now()
                hour = dt.hour
                if hour < 16:  # 额外对时间进行判断
                    return self.get_working_day_before_day(date.today())
            return last_date

    def recent_30_open_limit_down_to_file(self, up_down='up', days=30):
        """
        获得最近30天内涨/跌停的所有股票的信息，并计算涨/跌停后两天的股价变化情况。
        剔除开始统计的第一天和最后一天的涨/跌停股票，因为它们仍在涨/跌停过程中，统计的持续涨跌天数不全。
        倒数第二天的股票信息中，停止涨/跌停后第二天的增幅数据为'OutOfRange'，因为第二天的增幅需要明天的数据，
        明天还未到来，因此没有明天的数据，无法计算。
        增幅显示为NaN，则表示股票存在停牌。
        该方法主要为连跌后股票的增幅变化提供依据。
        返回的'连涨/跌出现次数'：1 表示正常情况，出现了一次连续涨/跌；2 表示出现了一次或多次连续涨跌；-2 表示以停牌持续到
        统计天数末尾；-3 表示倒数第二天的股票，这些股票不存在'连涨/跌后增幅2'。

        :param up_down: 统计连续涨停还是连续跌停，可以取值'up'或'down'，这里的涨/跌停不包括ST股。
        :param days: 统计的天数
        :return: codes, lastDays 股票代码和连续跌停的天数
        """
        # 确保OpenLimit文件夹存在
        file = os.path.join(self.data_folder, "OpenLimit")
        if not os.path.exists(file):
            os.mkdir(file)
        file = os.path.join(file, "recent30Down.csv")

        if up_down.lower() == 'up':
            title = "涨"
        elif up_down.lower() == 'down':
            title = "跌"
        else:
            logger.debug("参数未识别，up_down只能取值'up'或'down'！")

        # 拿到30天内每天的开盘涨/跌停股，以及对应的日期的列表
        limits, dates = self.get_open_limit_of_n_trading_days(days, up_down, sort='ascend')
        limits_set = []  # 将limit_downs中的每一个元素转化为集合
        for limit in limits:
            limits_set.append(set(limit))

        # <<<<<<<<<<<<<<<<<<<<<<首先删除数据不全的跌停股-----------------------------
        temp_set = limits_set[0]
        for i in range(1, len(limits_set)):
            """
            将temp_set和limits_set[i]的共有元素选出来，删除其他不共有的元素；因为如果一个元素在temp_set中，但
            不在limits_set[i]中，那么相对与limits_set[i]更靠内侧的日期的股票中再出现的temp_set中的元素就不属于数据
            不完整的股票了。举例如下：
            temp_set=['000001','000002']
            limits_set[0]=['000001','000002']
            limits_set[1]=['000001','000002','200002']
            limits_set[2]=['200003']
            limits_set[3]=['000001']
            limits_set[4]=['200005']
            则'000001'虽然在temp_set中，但是他在limits_set[3]中出现的时候，它的数据是全的。上一行是删除外侧日期
            即limits_set[0]和limits_set[1]中的'000001'，然后在limits_set[2]中没有'000001'，这是就要删除temp_set中的
            '000001'，下一行的工作就是删除temp_set中的'000001'
            """
            temp_set = limits_set[i] & temp_set
            # 从第一个集合里删除temp_set的元素，这里第一个集合指的是靠近最外边的日期的股票
            limits_set[i] = limits_set[i] - temp_set
            if len(temp_set) == 0:
                break
        # 删除另一侧的
        temp_set = limits_set[-1]
        for i in range(len(limits_set) - 2, 0, -1):
            temp_set = limits_set[i] & temp_set
            limits_set[i] = limits_set[i] - temp_set
            if len(temp_set) == 0:
                break
        # 删除边缘两天的所有股票，因为他们的持续天数都可能不全
        limits_set[0] = set()
        limits_set[-1] = set()
        # ----------------------首先删除数据不全的跌停股>>>>>>>>>>>>>>>>>>>>>>>>>>>>>

        # <<<<<<<<<<<<<<<<<<<<<<获得30天内所有出现过的开盘跌停股的集合-----------------
        codes = limits_set[0]  # 所有跌停股列表，为limit_downs_set所有元素的并集
        for i in range(1, len(limits_set)):
            codes = codes | limits_set[i]
        # ----------------------获得30天内所有出现过的开盘跌停股的集合>>>>>>>>>>>>>>>>>
        # 将limit_downs_set转换为列表，以方便进行索引
        limits = list(limits_set)
        # 将limits中每个元素转换为列表
        limits = [list(e) for e in limits]
        # 将codes转换为列表
        codes = list(codes)
        codes.sort(reverse=False)  # 随便排个序，不影响结果，只是为了方便每次调试时股票顺序一致好对比
        lastDays = []  # 每个涨/跌停股对应的持续天数
        increase_after1 = []  # 每个涨/跌停股停止涨/跌停后的第二天开盘价涨跌幅
        increase_after2 = []
        flags = []  # 标记当前股数据是否已经统计
        # -3表示倒数第二天的股票，这些股票的increase_after2不存在，因为涉及到了明天的数据，明天数据不存在，使得对应增幅无法计算
        # -2表示以跌停停牌，至今没有开牌
        # -1没统计，
        # 0正在统计/至今还处于跌停状态，
        # 1统计过一次，
        # 2开盘跌停在30天内出现一次以上

        for i in range(len(codes)):  # 数据齐全的所有股票代码
            lastDays.append(0)
            increase_after1.append(None)
            increase_after2.append(None)
            flags.append(-1)
        for i in range(len(codes)):  # 遍历所有需要统计的代码
            for j in range(len(limits)):  # 遍历所有日期的涨/跌停股
                if codes[i] in limits[j]:  # 首先查找包含codes[i]的日期，然后开始进入统计
                    if flags[i] == -1 or flags[i] == 0:
                        lastDays[i] = lastDays[i] + 1
                        flags[i] = 0
                    else:
                        flags[i] = flags[i] + 1
                else:
                    if flags[i] == 0:
                        flags[i] = 1
                        # 获得昨天和今天的股票数据，因为连续跌停时，停牌股概率大增，这里要考虑停牌影响
                        yesterday = self.get_data_of_day(codes[i], dates[j - 1], next_day=-1)
                        today = self.get_data_of_day(codes[i], dates[j], next_day=0)
                        if j < len(limits) - 1:
                            tomorrow = self.get_data_of_day(codes[i], dates[j + 1], next_day=0)
                        else:
                            tomorrow = []
                            outOfRange = True

                        if len(today) == 0:  # 因为dates中的日期都是工作日，如果查不到数据，则必然是股票停牌
                            flags[i] = -2
                        elif len(tomorrow) == 0:  # 停牌
                            open_yest = float(yesterday['close'])
                            open_today = float(today['close'])
                            increase_after1[i] = (open_today - open_yest) / open_yest
                            flags[i] = -2
                            if outOfRange:
                                increase_after2[i] = 'OutOfRange'
                                flags[i] = -3
                        else:
                            open_yest = float(yesterday['close'])
                            open_today = float(today['close'])
                            open_next = float(tomorrow['close'])
                            increase_after1[i] = (open_today - open_yest) / open_yest
                            increase_after2[i] = (open_next - open_today) / open_today
        # <<<<<<<<<<<<<<<<<<<<<<<获得具有完整跌停信息的股票的数据----------------------
        localData = {'股票代码': codes, '持续天数': lastDays, '连{}后增幅1'.format(title): increase_after1,
                     '连{}后增幅2'.format(title): increase_after2, '连{}出现次数'.format(title): flags}
        df = pd.DataFrame(localData)
        df.to_csv(file, encoding="utf8", index=False)
        logger.debug("{}到{}日连续{}停股票统计：\n{}".format(dates[0], dates[-1], title, df))
        return df

    def get_open_limit_and_lastdays(self, up_down='up', days: int = 10, last_date=date.today()):
        """
        将最近一个工作日的开盘涨/跌停股及其持续涨/跌停天数写到文件"./OpenLimit/recentDown.csv"

        :param up_down:
        :param days: 统计的天数，这里设置为10，因为连续涨/跌停
        :param last_date: 如果指定last_date，则获得该日期时所有开盘涨/跌停股的信息
        :return: DataFrame 股票代码和连续跌停的天数
        """
        # 处理up_down参数
        if up_down.lower() == 'up':
            title = "Up"
            title1 = "涨停"
        elif up_down.lower() == 'down':
            title = "Down"
            title1 = "跌停"
        else:
            logger.error("输入参数未识别，up_down只能取值'up'或'down'！")
        # 确保OpenLimit文件夹存在
        file = os.path.join(self.data_folder, "OpenLimit")
        if not os.path.exists(file):  # 确保文件夹存在
            os.mkdir(file)

        logger.debug("获取{}的{}股代码及持续天数...".format(last_date, title1))
        config.holdLoggingLevel(logging.WARN)
        limits, dates = self.get_open_limit_of_n_trading_days(days, up_down, 'descend', last_date=last_date)
        config.holdLoggingLevel('end')
        file = os.path.join(file, "recent{}{}.csv".format(title, dates[0]))
        if os.path.exists(file):
            df = pd.read_csv(file, encoding=self.encode, dtype={'股票代码': str})
        else:
            today_codes = limits[0]
            codes = today_codes
            lastDays = []
            flag = []
            for _ in codes:  # 初始化最近一天的涨/跌停股，每一个股对应的持续天数初始化为1
                lastDays.append(1)
                flag.append(True)
            for i in range(1, days):  # 天数
                for j in range(len(codes)):  # 第一天涨/跌股的数量，逐个检查每个股
                    if (codes[j] in limits[i]) and flag[j]:  # 如果前一天的跌停股列表里仍有code，则code的持续天数+1
                        lastDays[j] = lastDays[j] + 1
                    else:  # 如果前一天的跌停股列表里没有code，则将code标记为False，其lastdays列表中对应的天数将不再改变
                        flag[j] = False
            data = {'股票代码': codes, '持续天数': lastDays}
            df = pd.DataFrame(data)
            df.to_csv(file, encoding=self.encode, index=False)
        logger.debug("get_open_limit_and_lastdays()获得{}的{}股和持续天数：\n{}".format(last_date, title1, df))
        return df

    def get_open_limit_of_n_trading_days(self, days=30, up_down='up',
                                         sort='ascend', last_date=date.today()):
        """
        获得最近n个交易日的开盘涨/跌停股，包括今天（如果今天是交易日）；这里的涨/跌停不包括ST股。

        :param days: 统计的天数。
        :param up_down: 涨停还是跌停，取值：'up'或者'down'，默认是'up'。
        :param sort: 排序方式，'ascend'：从前到后，'descend'：从后到前；默认从前到后。
        :param last_date: 指定日期，如果不指定，则是最近的n个交易日，如果指定，则以指定日期为最后一个交易日。
        :return: limit_down_of_days和dates: [[],[],...[]]，列表中的每一个列表对应每一天的跌停股，result[0]是最近一个交易日的跌停股，如果今天是交易日，则result[0]就是今天。result[n]是第n个交易日前的跌停股，dates: 返回长度为n的date列表,列表为字符串，格式为"%Y-%m-%d"
        """
        if up_down.lower() == 'up':
            tmp = '涨停'
        elif up_down.lower() == 'down':
            tmp = '跌停'
        else:
            logger.error("参数未识别，up_down只能取值'up'或'down'！")
        config.loggingTitle("获取最近{}个交易日的{}信息".format(days, tmp), 'start')
        date_times = self.get_last_n_working_day(days, sort, last_date=last_date)
        open_limit_list = []

        for day_datetime in date_times:
            logger.debug("获取{}的开盘{}股...".format(day_datetime, tmp))
            config.holdLoggingLevel(logging.WARN)  # 临时将日志级别调高到警告，以取消局部的DEBUG日志输出
            limit = self.get_open_limit(up_down, day_datetime)
            open_limit_list.append(limit)
            config.holdLoggingLevel('end')  # 恢复日志输出级别
            logger.debug("{}的开盘{}股有：{}".format(day_datetime, tmp, limit))

        config.loggingTitle("获取最近{}个交易日的{}信息".format(days, tmp), 'end')
        return open_limit_list, date_times

    def get_working_day_before_day(self, day_datetime: datetime.date = date.today(), day_num: int = 1) -> datetime.date:
        """
        获得指定日期的前一个交易日或后一个交易日，需要注意的是，该函数不对指定日期是否为工作日进行判断，
        因此，获得最近一个交易日，不能使用get_working_day_before_day(date.today(),1)
        :param day_datetime: datetime类型，默认为今天
        :param day_num: 可取正整数和-1，前day_num个交易日，取-1时,表示指定日期的后一个交易日，不可以取其他负值
        :return: date_datetime】,-----返回对应交易日的datetime类型数据
        """
        # day = day_datetime.strftime("%Y-%m-%d")
        if day_num < -1:
            logger.error("day_num不能取-1以外的负值（{}）".format(day_num))
            return None
        if day_num == 1 or day_num == -1:
            one_day = datetime.timedelta(days=1)
            if day_num == -1:
                pre_day = day_datetime + one_day
            else:
                pre_day = day_datetime - one_day
            while self.is_holiday(pre_day):
                if day_num == -1:
                    pre_day = pre_day + one_day
                else:
                    pre_day = pre_day - one_day
            return pre_day
        else:  # day_num为大于1的整数
            day_datetime = self.get_working_day_before_day(day_datetime, 1)  # 计算上一个交易日
            return self.get_working_day_before_day(day_datetime, day_num - 1)  # 返回上一个交易日的上day_num-1个交易日

    @staticmethod
    def get_realtime_price(symbol):
        """
        获得股票实时价格
        只有实时价格需要实时获取，开盘、收盘等价格可能不需要实时更新，为了节省时间，顺便更新其他价格
        """
        data = ts.get_realtime_quotes(symbol)  # ts的实时行情接口在某些股票中会报错
        # ，因为某些股票的数据多了一个空数据，需要等待Tushare更新或者覆盖get_realtime_quotes中
        # 的相关参数，具体可以在get_realtime_quotes的下面代码之前，添加代码
        # df = pd.DataFrame(data_list, columns=ct.LIVE_DATA_COLS)
        # 添加的代码为：
        # if len(data_list[0])>33:
        #   data_list=data_list[0][:-1]
        name = str(data['name'][0])  # 这四条语句的执行是本地执行，且耗时很少
        open_price = float(data['open'])
        high = float(data['high'])
        low = float(data['low'])
        now_price = float(data['price'])
        return name, now_price, open_price, high, low

    def init_storage(self):
        """
        初始化存储的数据库操作对象
        """
        if self.storage == "Sqlite数据库":
            if self.engine is None:
                if self.project.db_name is None:
                    self.project.db_name = 'stocks.db'
                self.engine = create_engine(f'sqlite:///{self.project.db_name}', echo=False, encoding=self.encoding)
                self.sql = SqlOperator(self.engine)
                self.sql.create_all_base_table()
            else:
                pass
        elif self.storage == "Mysql数据库":
            if self.engine is None:
                user = self.project.db_user
                passwd = self.project.db_passwd
                ip = self.project.db_ip
                port = self.project.db_port
                db = self.project.db_name

                self.engine = create_engine(
                    # f"mysql+mysqlconnector://{user}:{passwd}@{ip}:{port}/{db}?charset={self.encoding}", echo=False,
                    f"mysql://{user}:{passwd}@{ip}:{port}/{db}?charset={self.encoding}", echo=False,
                )  # mysql数据库连接时charset最好在url中指定，因为SqlAlchemy2.0在外部指定时会报错
                self.sql = SqlOperator(self.engine)
                self.sql.create_all_base_table()
            else:
                pass



import torch
import torch.nn as nn
import random
import numpy as np
import yangke.stock.globalVar as gv
import os
import sys
import yangke.stock.dataset.data_reader as dr
import pandas as pd
from torch.utils.data import DataLoader, TensorDataset, Dataset
from yangke.common.config import holdLoggingLevel
import logging

torch.manual_seed(10000)
np.random.seed(10000)
random.seed(10000)


class MAXLoss(nn.Module):
    """
    自定义Loss函数，主要在于指定loss为批数据中最大的，torch自带的loss只能使用批数据的平均值
    """
    def __init__(self):
        super(MAXLoss, self).__init__()

    def forward(self, pred, target):
        ret = torch.abs(pred - target)
        ret = torch.pow(ret, 2)
        loss = ret.max()
        return loss


def prediction(stock_codes: list, method: str = "fit") -> pd.DataFrame:
    """
    预测列表中的股票涨幅，不会预测最近上市的股票——即股票上市时间小于(bptt_len+100)天的股票。

    新上市股票建议使用专用方法进行预测。

    :param stock_codes: 股票代码的列表
    :param method: 预测方法，目前支持'fit','classify','fit-1'
    :return: pd.DataFrame()，第一列为股票代码，第二列为涨幅
    """
    if method == "fit":
        method1 = lstm_stock_use_in_build_rnn_fit
    elif method == "classify":
        method1 = lstm_stock_classify
    elif method == "fit-1":
        method1 = lstm_stock_fit  # 所有股票共用同一个预测模型
    else:
        gv.logger.critical("method名不正确，接受到{}，但目前只支持{}".format(method, ['fit', 'classify', 'fit-1']))
        sys.exit(0)

    model_saved_folder = os.path.join(gv.dataDirectory, "nn-model/pytorch/lstm-{}".format(method))
    if not os.path.exists(model_saved_folder):
        os.makedirs(model_saved_folder)

    pred_list = []
    stock_codes_ = []  # 重新声明一个股票代码列表，用于记录可以正确预测涨跌幅的股票代码

    for i in range(len(stock_codes)):
        stock_code = stock_codes[i]
        model_saved_path = os.path.join(model_saved_folder, "{}.pth".format(stock_code))
        holdLoggingLevel(logging.WARNING)
        if '-1' in method:  # 如果是共用model进行预测，则传入公共model名
            pred = method1(stock_code, os.path.join(model_saved_folder, 'allInOneModel.pth'))
        else:
            pred = method1(stock_code, model_saved_path)
        holdLoggingLevel()
        gv.logger.debug("预测进度(完成预测的股票数)/(总股票数)：{}/{}，当前预测股票代码：{}".format(i, len(stock_codes), stock_code))
        if pred is not None:  # 如果有预测结果，则添加，有些没有结果的则不添加
            stock_codes_.append(stock_code)
            pred_list.append(pred)
        else:
            gv.logger.info("股票{}的预测结果为空，该股票可能是新股，数据量不足以进行预测！".format(stock_code))

    pred_dict = {'stock_code': stock_codes_, 'p_change': pred_list}
    result = pd.DataFrame.from_dict(pred_dict)
    return result


def get_data(stock_code: str or list, col_x=['close'], col_y=None, days=30, split=0.7):
    """
    对应每只股票获得其所有历史数据，组成训练集，测试集和需要预测的数据predX。

    可用于拟合模型和分类模型

    :param stock_code:
    :param col_x: 神经网络输入的数据列
    :param col_y: 神经网络输出的数据列
    :param days: 使用前days天的数据预测后一天的数据
    :param split: 数据集分割比例，如训练数据：验证数据 = 7:3，则split=0.7
    :return:
    """
    batch_size = 8
    result = dr.get_data_days_list(stock_code, col_x=col_x, col_y=col_y, days=days,
                                   normalization='z-score', need_pred=True,
                                   split=split)
    if result is None:  # 有些新开股票，数据太少不足以生成一条训练数据，则返回结果是None
        return None
    train_x, train_y, test_x, test_y, predict_x = result

    if col_y is None:
        train_x, train_y = torch.from_numpy(train_x).type(torch.float), torch.from_numpy(train_y).type(torch.float)
        test_x, test_y = torch.from_numpy(test_x).type(torch.float), torch.from_numpy(test_y).type(torch.float)
    elif isinstance(col_y, list):  # 如果是分类问题，则y是整数，表示类别
        train_x, train_y = torch.from_numpy(train_x).type(torch.float), torch.from_numpy(train_y).type(torch.long)
        test_x, test_y = torch.from_numpy(test_x).type(torch.float), torch.from_numpy(test_y).type(torch.long)
    predict_x = torch.from_numpy(predict_x).type(torch.float)

    train_dataset = TensorDataset(train_x, train_y)
    test_dataset = TensorDataset(test_x, test_y)
    # 只有在数据量很大的时候使用多核加载才能提高速度，数据量小时反而会降低速度
    train_loader = DataLoader(dataset=train_dataset, batch_size=batch_size, shuffle=True)  # , num_workers=2)
    test_loader = DataLoader(dataset=test_dataset, batch_size=batch_size, shuffle=True)  # , num_workers=2)
    if isinstance(col_y, list):  # 如果是分类问题
        num_classes = len(col_y) + 1
        return train_loader, test_loader, batch_size, days, num_classes, predict_x
    else:  # 如果是拟合问题
        return train_loader, test_loader, batch_size, days, predict_x


def need_update_model(model):
    """
    判断一个pytorch模型是否需要更新

    :param model:
    :return:
    """
    if os.path.exists(model):
        return False
    return True


class RNN_fit(nn.Module):
    # bptt_len=30，即过去30天的数据，input_size=1即每天取一个数据
    def __init__(self, bptt_len, input_size, hidden_size, num_layers):
        super(RNN_fit, self).__init__()
        # 定义模型中的神经网络层
        self.rnn = nn.LSTM(input_size=input_size, hidden_size=hidden_size, num_layers=num_layers, batch_first=True)

        # 用于将隐含层的输入映射到输出
        self.linear = nn.Linear(in_features=hidden_size * bptt_len, out_features=1)

    # input: [8, 30, 1], hidden: [batch_size, num_layers, hidden_size] # 这里把batch_size调到第一位
    def forward(self, x):
        # rnn层的输入和输出形状是相同的
        output, _ = self.rnn(x)

        # 需要调用output.contiguous()将Tensor的存储整理为连续存储，否则可能报错，原因可能是多卡存储或其他
        output = output.contiguous().view(output.size(0), output.size(1) * output.size(2))  # [8, 30, 1]->[8, 30]

        # 添加一个激活函数，该层为非线性，使得网络可以处理非线性关系，LSTM中有非线性，可能不需要，这里对比下训练结果
        # 不添加激活函数 # 0.20275171101093292, test loss 4.4828590740358205, adjust lr to 0.000625
        output = torch.sigmoid(output)  # loss 0.00398010341450572, test loss 4.799441468508946
        # output = torch.tanh(output)  # loss 0.209489107131958, test loss 4.479561124922912, adjust lr to 0.000625
        # output = torch.relu(output)  # loss 1.0070337057113647, test loss 4.457948378744842
        # output = torch.nn.functional.softplus(output)  # loss 0.0015654880553483963, test loss 4.68150989030827

        # 这里将lstm层输出的Tensor形状的前两位合并，只有输出的维度是需要传出去，以和labels进行比较计算loss
        output = self.linear(output)  # 将[*, input_size]映射到[*, output_size], [8, 30]->[8, 1]

        output = output.squeeze(-1)  # [8, 1] -> [8]
        return output

    def prediction(self, x):
        with torch.no_grad():
            if x.dim() == 2:  # 说明只有一个数据，则添加batch对应的维度
                x = x.unsqueeze(0)
            output = self.forward(x)
        return output.parent()


class RNN_classify(nn.Module):
    # bptt_len=30，即过去30天的数据，input_size=1即每天取一个数据, num_classes即分类数量
    def __init__(self, bptt_len, input_size, hidden_size, num_layers, num_classes):
        super(RNN_classify, self).__init__()
        # 定义模型中的神经网络层
        self.rnn = nn.LSTM(input_size=input_size, hidden_size=hidden_size, num_layers=num_layers, batch_first=True)
        # 用于将隐含层的输入映射到输出
        self.linear = nn.Linear(in_features=hidden_size * bptt_len, out_features=num_classes)

    # input: [8, 30, 1], hidden: [batch_size, num_layers, hidden_size] # 这里把batch_size调到第一位
    def forward(self, x):
        # rnn层的输入和输出形状是相同的
        output, _ = self.rnn(x)

        # 需要调用output.contiguous()将Tensor的存储整理为连续存储，否则可能报错，原因可能是多卡存储或其他
        output = output.contiguous().view(output.size(0), output.size(1) * output.size(2))  # [8, 30, 1]->[8, 30]

        # 添加一个激活函数，该层为非线性，使得网络可以处理非线性关系
        # output = torch.sigmoid(output)  # loss 0.5425181984901428, test loss 0.7410567089312339, adjust lr to 0.000625
        # output = torch.tanh(output)  # loss 0.19794771075248718, test loss 0.6420857055338821
        output = torch.relu(output)  # loss 0.12568356096744537, test loss 0.6467462378430229
        # output = torch.nn.functional.softplus(output)  # loss 0.5891538262367249, test loss 0.7501011504603259, adjust lr to 0.000625

        # 这里将lstm层输出的Tensor形状的前两位合并，只有输出的维度是需要传出去，以和labels进行比较计算loss
        output = self.linear(output)  # 将[*, input_size]映射到[*, output_size], [8, 30]->[8, hidden_size]

        return output

    def prediction(self, x):
        with torch.no_grad():
            if x.dim() == 2:
                x = x.unsqueeze(0)  # 增加一维作为batch_size为1的数据，如果batch_first=False，则x.unsqueeze(1)
            output = self.forward(x).squeeze(0)
            sm = nn.Softmax(dim=0)
            output = sm(output).numpy()
            classes = output.argmax()  # 获得最大元素的索引，是为分类号
        return classes, output


def evaluate(model: RNN_fit, test_loader, loss_fn):
    model.eval()
    total_loss = 0.
    total_count = 0.
    with torch.no_grad():
        for i, (data, labels) in enumerate(test_loader):
            output = model(data)

            loss = loss_fn(output, labels)
            total_count += len(data)  # 一共多少数据
            total_loss += loss.parent() * len(data)  # 总的loss
    loss = total_loss / total_count
    model.train()
    return loss


def train(model, epochs, train_loader, test_loader, optimizer, scheduler, loss_fn, model_path):
    train_losses = []
    val_losses = []  # 记录训练过程中的losses，方便后续绘图
    flag = 0
    for epoch in range(epochs):
        # input: [batch_size, bptt_len, input_size], labels: [batch_size]
        for i, (data, labels) in enumerate(train_loader):
            # 输入形状是[8, 30, 1]，即[batch_size, bptt_len, input_size]
            out = model(data)  # out: [8]

            loss = loss_fn(out, labels)
            optimizer.zero_grad()
            loss.backward()
            # torch.nn.utils.clip_grad_norm_(model.parameters(), 1)  # 防止参数过大
            optimizer.step()
            if i % 10 == 0:
                val_loss = evaluate(model, test_loader, loss_fn)

                if len(val_losses) == 0 or val_loss < min(val_losses):
                    gv.logger.info("epoch {}, step {}, loss {}, test loss {}".format(epoch, i, loss, val_loss))
                    torch.save(model.state_dict(), model_path)
                elif loss < min(train_losses) and flag < 3:  # 训练集损失减小，测试集损失增大，继续训练，
                    flag = flag + 1  # 连续三次如此，则更改学习速率
                    gv.logger.info("epoch {}, step {}, loss {}, test loss {}".format(epoch, i, loss, val_loss))
                else:  # 表示训练数据和测试数据都变差
                    flag = 0
                    scheduler.step()  # 会自动把optimizer中的lr修改为更小
                    real_time_lr = optimizer.param_groups[0]['lr']
                    gv.logger.info("epoch {}, step {}, loss {}, test loss {}, adjust lr to {}"
                                   .format(epoch, i, loss, val_loss, real_time_lr))

                # 记录训练和测试数据集的损失
                train_losses.append(loss)
                val_losses.append(val_loss)
    return model


def lstm_stock_use_in_build_rnn_fit(stock_code: str, model_path: str = None):
    """
    使用torch内置的LSTM层进行股票预测，与lstm_stock_use_in_build_rnn()功能相同，只是为了保证复现，修改了一些随机数的机制

    :param stock_code:
    :param model_path: 神经网络模型的保存路径，最好是绝对路径，否则会保存到主程序的目录下
    :return:
    """
    stock_code = str(stock_code)
    if model_path is None:
        os.makedirs(f"{gv.dataDirectory}/torch/model/", exist_ok=True)  # 确保路径存在
        saved_model_pth = f"{gv.dataDirectory}/torch/model/{stock_code}.pth"
    else:
        saved_model_pth = model_path

    col_x = ['close', 'volume']
    result = get_data(stock_code, col_x=col_x)
    if result is None:  # 没有股票数据
        return None
    train_loader, test_loader, batch_size, days, predX = result

    model = RNN_fit(bptt_len=days, input_size=len(col_x), hidden_size=64, num_layers=2)
    if not need_update_model(saved_model_pth):  # 如果不需要更新模型
        model.load_state_dict(torch.load(saved_model_pth))
        model.eval()
        predY = model.prediction(predX)
        return predY

    learning_rate = 0.01
    optimizer = torch.optim.Adam(model.parameters(), lr=learning_rate)
    scheduler = torch.optim.lr_scheduler.ExponentialLR(optimizer, 0.5)

    # loss_fn = torch.nn.MSELoss()
    loss_fn = MAXLoss()  # 因为我们主要关心最差的时候别太差

    model = train(model, 2, train_loader, test_loader, optimizer, scheduler, loss_fn, saved_model_pth)

    pred = model.prediction(predX)
    gv.logger.info("预测股票{}明天的涨幅为{}".format(stock_code, pred))
    return pred


def lstm_stock_classify(stock_code: str, model_path: str = None):
    stock_code = str(stock_code)
    if model_path is None:
        saved_model_pth = "{}.pth".format(stock_code)
    else:
        saved_model_pth = model_path

    # 输入数据的列数，假如考虑过去30天开盘价和收盘价，则input_size=2, bptt_len=30
    def main(input_size):
        EPOCHS = 2

        model = RNN_classify(bptt_len, input_size, hidden_size=64, num_layers=2, num_classes=num_classes)

        if not need_update_model(saved_model_pth):  # 如果不需要更新模型
            model.load_state_dict(torch.load(saved_model_pth))
            model.eval()
            predY = model.prediction(predX)
            return predY

        learning_rate = 0.01
        optimizer = torch.optim.Adam(model.parameters(), lr=learning_rate)
        scheduler = torch.optim.lr_scheduler.ExponentialLR(optimizer, 0.5)

        loss_fn = torch.nn.CrossEntropyLoss()

        train_losses = []
        val_losses = []  # 记录训练过程中的losses，方便后续绘图
        flag = 0
        flag1 = 0
        for epoch in range(EPOCHS):
            # input: [batch_size, bptt_len, input_size], labels: [batch_size]
            for i, (data, labels) in enumerate(train_loader):
                # 输入形状是[8, 30, 1]，即[batch_size, bptt_len, input_size]
                out = model(data)  # out: [batch_size, num_classes]:[8, 3]

                loss = loss_fn(out, labels)  # out: [8, 3], labels: [8]
                optimizer.zero_grad()
                loss.backward()
                # torch.nn.utils.clip_grad_norm_(model.parameters(), 1)  # 防止参数过大
                optimizer.step()
                if i % 10 == 0:
                    val_loss = evaluate(model, test_loader, loss_fn)

                    if len(val_losses) == 0 or val_loss < min(val_losses):
                        gv.logger.info("epoch {}, step {}, loss {}, test loss {}".format(epoch, i, loss, val_loss))
                        torch.save(model.state_dict(), saved_model_pth)
                    elif loss < min(train_losses) and flag < 3:  # 训练集损失减小，测试集损失增大，继续训练，
                        flag = flag + 1  # 连续三次如此，则更改学习速率
                        gv.logger.info("epoch {}, step {}, loss {}, test loss {}".format(epoch, i, loss, val_loss))
                    elif flag < 3:  # 表示训练数据和测试数据都变差
                        flag = 0
                        flag1 = flag1 + 1
                        scheduler.step()  # 会自动把optimizer中的lr修改为更小
                        real_time_lr = optimizer.param_groups[0]['lr']
                        gv.logger.info("epoch {}, step {}, loss {}, test loss {}, adjust lr to {}"
                                       .format(epoch, i, loss, val_loss, real_time_lr))

                    # 记录训练和测试数据集的损失
                    train_losses.append(loss)
                    val_losses.append(val_loss)

        # predY=0表示涨幅小于1%，predY=1：涨幅[1%, 5%)，predY=2：涨幅[5%,)
        predY, probability = model.prediction(predX)  # 获得分类号，model(x)获得的是最后一个线性层的输出
        return predY, probability

    col_x = ['close', 'volume']
    col_y = [1, 5]
    result = get_data(stock_code, col_x=col_x, col_y=col_y, days=30)
    if result is None:  # 没有股票数据
        return None
    train_loader, test_loader, batch_size, bptt_len, num_classes, predX = result
    pred, prob = main(len(col_x))
    gv.logger.info("预测股票{}的明天涨幅为{}，概率分布为{}".format(stock_code, pred, prob))
    return pred, prob.max()


def lstm_stock_fit(stock_code, model_path: str = None):
    """
    将所有股票数据用同一个模型拟合

    :param stock_code:
    :param model_path:
    :return:
    """
    if model_path is None:
        saved_model_pth = "allInOneModel.pth"
        saved_model_pth = os.path.abspath(saved_model_pth)
        gv.logger.warning('No model path is specified, The model is saved in {}'.format(saved_model_pth))
    else:
        saved_model_pth = model_path

    col_x = ['close']
    col_y = None
    days, split = 30, 0.7
    input_size = len(col_x)
    epochs = 2
    learning_rate = 0.01
    # 根据股票代码获得predX
    result = dr.get_data_days_list(stock_code, col_x=col_x, col_y=col_y, days=days,
                                   normalization='z-score', need_pred=True,
                                   split=split)
    if result is None:  # 有些新开股票，数据太少不足以生成一条训练数据，则返回结果是None
        return None
    _, _, _, _, predX = result
    predX = torch.from_numpy(predX).type(torch.float)

    # 无论如何，使用全部股票数据进行模型训练
    stock_codes = gv.tsd.download_all_stocks()
    train_codes, _, test_codes, _ = dr.__split_dataset(stock_codes, stock_codes, split)
    dataset_train = dr.StockDataset(train_codes)
    dataset_test = dr.StockDataset(test_codes)
    train_loader = DataLoader(dataset=dataset_train, batch_size=8, shuffle=True)
    test_loader = DataLoader(dataset=dataset_test, batch_size=8, shuffle=False)

    model = RNN_fit(bptt_len=days, input_size=input_size, hidden_size=64, num_layers=2)
    optimizer = torch.optim.Adam(model.parameters(), lr=learning_rate)
    scheduler = torch.optim.lr_scheduler.ExponentialLR(optimizer, 0.5)

    # loss_fn = torch.nn.MSELoss()
    loss_fn = MAXLoss()  # 因为我们主要关心最差的时候别太差
    model = train(model, epochs=epochs, train_loader=train_loader, test_loader=test_loader,
                  optimizer=optimizer, scheduler=scheduler, loss_fn=loss_fn, model_path=saved_model_pth)

    result = model.prediction(predX)
    return result



"""
定义操作流程
"""
import copy
import os.path
import time

from yangke.base import pic2ndarray
from yangke.common.win.win_x64 import *
from yangke.objDetect.ocr import ocr


class State:
    def __init__(self, hwnd, image_folder):
        self.hwnd = hwnd
        self.target = None
        self.type = "image"
        self.image_folder = image_folder
        self.state = "exist"  # 有"exist","absent","unchanging", "changing"四种取值，分别表示存在，不存在，保持不变，变化中
        self.threshold = 0.8
        self.pos = None
        self.section = None

    def set_state(self, target, state="exist", threshold=0.8, section=None):
        """

        :param target:
        :param state: 有"exist","absent","unchanging", "changing"四种取值，分别表示存在，不存在，保持不变，变化中
        :param threshold:
        :param section:
        :return:
        """
        self.type = "image"  # 状态的类型，是发现什么图片还是找到什么文字
        if isinstance(target, str):
            if os.path.exists(os.path.join(self.image_folder, target)):  # 首先尝试在指定的image_folder目录下查找目标图片
                self.target = os.path.join(self.image_folder, target)
            elif os.path.exists(target):  # 如果上面目录没查到，则在当前程序娙目录中查找
                self.target = os.path.abspath(target)
            try:
                self.target = pic2ndarray(self.target)
            except:
                self.type = "text"  # 如果target转换为图片时发生错误，则类型为text
        if target == 1:  # 当state为unchanging或changing时，target标记目标的类型，1表示图片不变，2表示文字
            self.type = "image"
        elif target == 2:
            self.type = "text"

        self.state = state  # 判断的条件是存在还是不存在，默认target存在则条件满足，返回True
        self.threshold = threshold
        self.section = section
        return copy.deepcopy(self)

    def is_state(self):
        if self.section is not None:
            _ = self.section
        else:
            _ = (0, 0, 0, 0)
        capture_pic_undecorated(self.hwnd, x1=_[0], y1=_[1], x2=_[2], y2=_[3])
        if self.type == "image":
            exists, pos = find_pic("temp.png", self.target, threshold=self.threshold)
            if exists:
                self.pos = pos
                if self.state == "exist":  # 要求存在，并且存在，则返回True
                    return True
                elif self.state == "absent":  # 要求不存在，但是存在，则返回False
                    return False
            else:
                if self.state == "exist":  # 要求存在，但是不存在，则返回False
                    return False
                elif self.state == "absent":  # 要求不存在，并且不存在，则返回True
                    return True
        else:
            text = ocr("temp.png", self.threshold, paragraph=True, method="paddleocr")
            if self.state == "exist":
                if self.target in text:
                    return True
                else:
                    return False
            elif self.state == "absent":
                if self.target in text:
                    return True
                else:
                    return False
            elif self.state == "unchanging":
                time.sleep(1)
                text1 = ocr("temp.png", self.threshold, paragraph=True, method="paddleocr")
                if text1 == text:  # 如果文字未改变，则返回True
                    return True
                else:
                    return False
            elif self.state == "changing":
                time.sleep(1)
                text1 = ocr("temp.png", self.threshold, paragraph=True, method="paddleocr")
                if text1 == text:
                    return False
                else:
                    return True
        return False


class Action:
    def __init__(self, hwnd, image_folder=""):
        self.hwnd = hwnd
        self.image_folder = image_folder
        self.act = "mouse"
        self.pos = None
        self.key = None
        self.right = None

    def set_action(self, pos=None, key=None, right=False):
        """
        设置动作，不支持鼠标键盘同时操作

        :param pos: 位置
        :param key: 按键，可取值：‘ctrl+f’,'shift+1'等
        :param right: 是否鼠标右键
        :return:
        """
        self.pos = pos
        self.key = key
        if self.key is not None:
            self.act = "key"
        else:
            self.act = "mouse"
        self.right = False
        return copy.deepcopy(self)

    def do(self):
        if self.act == "mouse":
            _ = self.pos
            do_click(_[0], _[1], self.hwnd, self.right)
        else:
            keys = self.key.strip().split("+")
            shift = []
            key = 0
            for k in keys:
                if len(k) == 1:
                    key = ord(k)
                elif k.lower() == "ctrl":
                    shift.append(win32con.VK_CONTROL)
                elif k.lower() == "alt":
                    shift.append(win32con.VK_MENU)
                elif k.lower() == "shift":
                    shift.append(win32con.VK_SHIFT)
                elif k.lower() == "esc":
                    key = win32con.VK_ESCAPE
            post_key(hwnd=self.hwnd, key=key, shift=shift)


class ActionNode:
    def __init__(self, in_node, action=None, out_node=None, else_action=None):
        self.condition: State = in_node  # 当前发现什么，即如果该状态满足，才进行相应操作
        self.action: Action = action  # 状态满足后需要执行的操作
        self.else_action = else_action  # 万一第一个状态不满足，需要执行的操作
        self.result: State = out_node  # 操作执行后，应当产生的结果，如果该结果没有发生，则再次执行self.action
        self.pos = self.condition.pos

    def do(self):
        _ = 0
        while not self.condition.is_state():  # 当不满足输入状态时，等待
            time.sleep(1)
            _ += 1
            if _ > 5 and self.else_action is not None:  # 当等待超过10秒时，尝试补救操作
                self.else_action.do()
            if _ > 10:  # 仍然操作无效时，报错
                logger.error("尝试了各种方法，但一直无法满足ActionNode的初始状态!")
                break
        if self.action is None:
            # 如果没有指定动作，则默认点击condition的坐标位置，因为condition一般为查找某个图标，该图标往往是点击目标
            self.action = Action(self.condition.hwnd, self.condition.image_folder)
            self.action.set_action(pos=(self.condition.pos[0], self.condition.pos[1]))
        self.action.do()
        time.sleep(1)
        if self.result is None:
            return True
        while not self.result.is_state():
            self.action.do()
            time.sleep(1)
        return True

# state = State(hwnd, image_folder="")
#
# action1 = ActionNode(state.set_state("*.png"), Action("Mouse", pos=(0, 0)),
#                      state.set_state("接受", section=(0, 0, 200, 200)))
# action1.do()



import time
import traceback

import selenium.common.exceptions as exceptions
from selenium.common import NoSuchElementException
from selenium.webdriver.edge.service import Service

from yangke.common.config import logger
from selenium import webdriver
from selenium.webdriver.edge.webdriver import WebDriver
from selenium.webdriver.common.by import By
from selenium.webdriver.common.keys import Keys
from selenium.webdriver.support.wait import WebDriverWait
from selenium.webdriver.support import expected_conditions as EC
from selenium.webdriver.common.action_chains import ActionChains
from yangke.common.fileOperate import write_as_pickle, read_from_pickle
from yangke.base import sound_beep

i = 1  # 开始下载的文件的编号
num_of_reports_in_single_page = 20  # 每一页的报告数量


def double_click_element(ele):
    action.double_click(ele).perform()


def switch_to_window_by_title(driver, title):
    for handle in driver.window_handles:
        driver.switch_to.window(handle)
        if title in driver.title:
            break
    return


def 打开技术报告页面(driver):
    driver.get("http://10.219.2.75/seeyon/main.do?method=main")  # 打开系统主页
    # driver.get("http://10.219.2.67/WebCenter/Open/00000000-0000-0000-0000-00000000000a")
    logger.debug(f"登录协同并打开技术报告页面...")
    element = WebDriverWait(driver, 10).until(
        EC.presence_of_element_located((By.XPATH, '//*[@id="loginBtn"]/div/span'))
    )
    time.sleep(1)
    element.click()
    driver.switch_to.frame("loginPreBtnIframeId")  # 切换到iframe中，iframe中的元素无法直接定位
    # 逆操作为：driver.switch_to.default_content()
    driver.find_element(by=By.XPATH, value='//*[@id="login_username"]').send_keys("luoliujuan")  # 输入账号名
    driver.find_element(by=By.XPATH, value='//*[@id="login_password"]').send_keys("Luo012210!")  # 输入密码
    driver.find_element(by=By.XPATH, value='//*[@id="login_button"]').click()  # 点击登录按钮
    element = WebDriverWait(driver, 10).until(
        EC.presence_of_element_located((By.XPATH, '//ul[@id="showedNav"]'))
    )  # 登录后等待界面加载
    element = element.find_element(By.XPATH, value='//li[contains(@title, "项目管理")]')  # 获取title中包含"项目管理"的元素
    main_window_handle = driver.current_window_handle  # 协同主页窗口
    element.click()  # 点击项目管理
    driver.implicitly_wait(10)  # 最长等待10s等待加载完成
    driver.switch_to.window(driver.window_handles[-1])  # 切换到最新打开的窗口
    element = WebDriverWait(driver, 10).until(
        EC.presence_of_element_located((By.XPATH, '//span[contains(text(), "科技成果")]'))
    )
    element.click()  # 点击科技成果
    if driver.find_element(By.XPATH, '//span[contains(text(), "技术报告")]').text == "":  # 如果页面中没有技术报告标签
        element = driver.find_element(By.XPATH, '//span[text()="报告"]')  # 则点击报告，展开报告项
        element.click()  # 展开报告
    element = driver.find_element(By.XPATH, '//span[contains(text(), "技术报告")]')
    element.click()
    time.sleep(10)
    # ------------------------------ 切换到技术报告页面中的子iframe中 -----------------------------------
    # 技术报告显示在新的iframe中
    driver.switch_to.frame(0)  # 页面中只有一个frame
    element = WebDriverWait(driver, 10).until(
        EC.presence_of_element_located((By.XPATH, '//div[contains(text(), "技术报告名称")]'))
    )
    # ------------------------------ 切换到技术报告页面中的子iframe中 -----------------------------------
    return driver, main_window_handle


def download_one_page_file(driver, start_file_id=1):
    """
    下载主页->科技成果->技术报告中主界面的iframe中一页的所有附件，调用该方法时，driver必须已经指向页面的iframe中

    :param driver:
    :return:
    """
    global i
    for j in range(1, num_of_reports_in_single_page + 1):  # 一页下载15个
        if i < start_file_id:
            i += 1
            continue
        try:
            ele = driver.find_element(By.XPATH, f'//div[text()="{i}"]')  # 文件编号是总的编号
            double_click_element(ele)  # 双击技术报告列表中的行，弹出技术报告框
        except NoSuchElementException:
            break
        time.sleep(4)
        driver.switch_to.default_content()
        ele_附件 = WebDriverWait(driver, 30).until(
            EC.presence_of_element_located((By.XPATH, '//iframe[contains(@name, "mini-iframe-")]'))
        )
        logger.debug(f"正在下载第{i}个文件")
        driver.switch_to.frame(ele_附件)  # 切换到弹出的技术报告对话框
        ele_附件 = WebDriverWait(driver, 10).until(
            EC.presence_of_element_located((By.XPATH, '//span[contains(text(), "附件")]'))
        )
        ele_附件.click()
        time.sleep(4)

        # 找到下一层iframe，这一层frame的name可能是"mini-iframe-120"或"mini-iframe-119"
        ele_frame = driver.find_elements(By.XPATH, '//iframe[contains(@name, "mini-iframe-")]')
        if len(ele_frame) == 1:
            ele_frame = ele_frame[0]
        else:
            logger.error(f"当前iframe下发现多个iframe，请处理！")
        driver.switch_to.frame(ele_frame)  # 切换到技术报告对话框右侧的表格中
        time.sleep(1)
        存在附件 = True
        try:
            WebDriverWait(driver, 10).until(
                EC.presence_of_element_located((By.XPATH, '//td[@id="1$cell$4"]'))
            )
            time.sleep(1)
        except exceptions.TimeoutException:
            存在附件 = False
            logger.debug(f"第{i}个项目没有附件")
        if 存在附件:
            # ---------------------------- 选中需要下载的文件 ------------------------------
            ele_pdf = driver.find_element(By.XPATH, "//span[@id='mini-5checkall']")  # 应该点击选择所有的按钮
            ele_pdf.click()
            time.sleep(1)
            # # ---------------------------- 选中需要下载的文件 ------------------------------
            # # ---------------------------- 点击下载按钮 ------------------------------
            ele_下载按钮 = driver.find_element(By.XPATH, '//button[@id="btnDownLoad"]')
            ele_下载按钮.click()
            # ---------------------------- 点击下载按钮 ------------------------------
        # ---------------------------- 点击关闭按钮 ------------------------------
        # 无论是否存在附件，都需要点击关闭按钮
        driver.switch_to.default_content()  # 切换到最外层的主页面所在的内容
        time.sleep(1)
        driver.switch_to.frame(
            driver.find_element(By.XPATH, '//iframe[contains(@name, "mini-iframe-")]'))  # 切换到中间层的iframe
        time.sleep(1)
        ele_close = driver.find_element(By.XPATH, '//a[@id="RGY_MOST_TechReport.CloseForm"]')
        ele_close.click()
        time.sleep(2)
        # ---------------------------- 点击关闭按钮 ------------------------------
        # ---------------------------- 解析页面切换回主界面中的项目列表中 ----------------------------------
        driver.switch_to.frame(0)
        # ---------------------------- 解析页面切换回主界面中的项目列表中 ----------------------------------
        i = i + 1


def next_page(driver: WebDriver):
    logger.info("点击下一页按钮")
    ele_next_page = WebDriverWait(driver, 10).until(
        EC.presence_of_element_located(
            (By.XPATH, '//span[@class="mini-button-icon mini-icon mini-iconfont mini-pager-next"]'))
    )
    ele_next_page.click()
    time.sleep(4)


def goto_page(driver: WebDriver, page):
    """
    翻到第page页

    :param driver:
    :param page:
    :return:
    """
    logger.debug(f"翻到第{page}页")
    ele = driver.find_element(By.XPATH, '//input[@type="text" and @class="mini-pager-num"]')
    double_click_element(ele)
    ele.send_keys(str(page))
    ele.send_keys(Keys.ENTER)


def download_file(driver):
    global i
    i = read_from_pickle("default.pkl") or i
    page = int((i - 0.5) // num_of_reports_in_single_page) + 1
    if page > 2:
        goto_page(driver, page)
    while True:  # 当有下一页内容时，则一直下载，直到找不到下一页按钮
        download_one_page_file(driver, i)
        next_page(driver)
        logger.debug("翻到下一页")


action = None


def run():
    global action, i
    driver: WebDriver = webdriver.Edge(service=Service(r"D:\msedgedriver.exe"))
    action = ActionChains(driver)  # 初始化一个动作对象
    try:
        driver, main_window = 打开技术报告页面(driver)
        download_file(driver)
        msg = "下载结束"
    except exceptions.TimeoutException or exceptions.ElementClickInterceptedException or exceptions.StaleElementReferenceException:
        msg = "下载超时"
    except exceptions.NoSuchElementException:
        i = 2  # 当前层级下载结束，重置索引序号
        sound_beep(500, 1000)
        msg = "下载结束"
    except:
        sound_beep(5000, 1000, 1000, 3)  # 下载出错提示音
        msg = "未知错误"
    finally:
        write_as_pickle(file="default.pkl", obj=i - 1)
        traceback.print_exc()
        driver.close()
        return msg


if __name__ == "__main__":
    # 重新下载需要删除掉运行目录下的default.pkl文件，否则会跳过default.pkl文件中记录的已经下载的文件编号
    res = run()
    while res == "下载超时" or res == "未知错误":  # 如果是下载超时，则重新尝试
        res = run()
        traceback.print_exc()
        logger.debug(f"下载过程中发生错误:{res}，尝试重新下载...")
